/*
 * $Id: Component.cc,v 1.4 2003/03/27 03:31:09 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "qualresUtility.h"
#include "Component.h"

gboolean componentProjectIterForeachFunc(GtkTreeModel *,
					 GtkTreePath *,
					 GtkTreeIter *,
					 gpointer );

/** Component constructor.
 */
Component::Component()
{
  selectDocWindow = NULL;
  selectCatWindow = NULL;
  selectSetWindow = NULL;
  projectDocMap = NULL;
  projectCatMap = NULL;
  projectSetMap = NULL;
}

/** Load a map.

    \param cMap is the component map

    \param type is the type of the Project Component
*/
void
Component::loadMap(map<const string, Component *> *cMap,
		   QualResObject::Type type)
{
  switch (type)
    {
    case QualResObject::DOCUMENT:
      projectDocMap = (map<const string, Document *> *) cMap;
      break;
    case QualResObject::CATEGORY:
      projectCatMap = (map<const string, Category *> *) cMap;
      break;
    case QualResObject::SET:
      projectSetMap = (map<const string, Set *> *) cMap;
      break;
    default:
      g_assert_not_reached();
      return;
    }
}

/** Load a model.

    \param cModel is the component model

    \param type is the type of the Project Component
*/
void
Component::loadModel(GtkTreeStore *cModel, QualResObject::Type type)
{
  switch (type)
    {
    case QualResObject::DOCUMENT:
      selectDocStore = cModel;
      break;
    case QualResObject::CATEGORY:
      selectCatStore = cModel;
      break;
    case QualResObject::SET:
      selectSetStore = cModel;
      break;
    default:
      g_assert_not_reached();
      return;
    }
}

/** Hide all the windows to select a Component.

*/
void
Component::hideAllSelectWindows()
{
  hideSelectWindow(QualResObject::DOCUMENT);
  hideSelectWindow(QualResObject::CATEGORY);
  hideSelectWindow(QualResObject::SET);
}

/** Hide the window to select a Component.

    \param type is the Component the select window to hide is
    associated with

*/
void
Component::hideSelectWindow(enum QualResObject::Type type)
{
  switch (type)
    {
    case QualResObject::DOCUMENT:
      if (selectDocWindow != NULL)
	gtk_widget_hide(selectDocWindow);
      break;
    case QualResObject::CATEGORY:
      if (selectCatWindow != NULL)
	gtk_widget_hide(selectCatWindow);
      break;
    case QualResObject::SET:
      if (selectSetWindow != NULL)
	gtk_widget_hide(selectSetWindow);
      break;
    default:
      g_assert_not_reached();
      return;
    }
}

/** Commit the pending new name.

    pendingSetName must have been called.  If the new name is
    different, the object is marked as changed and the window title is
    changed.
*/
void
Component::commitName()
{
  QualResObject::commitName();

  if (selectDocWindow != NULL)
      gtk_window_set_title(GTK_WINDOW(selectDocWindow),
			   (getPrettyName()).c_str());
  if (selectCatWindow != NULL)
      gtk_window_set_title(GTK_WINDOW(selectCatWindow),
			   (getPrettyName()).c_str());
  if (selectSetWindow != NULL)
      gtk_window_set_title(GTK_WINDOW(selectSetWindow),
			   (getPrettyName()).c_str());
}

/** Structure to hold the context for finding the iterator of a Component.
*/
struct ComponentProjectIterContext
{
  string name; ///< Name of the Component whose iterator we're searching for
  GtkTreeIter *iter; ///< Iterator of the Component we're searching for
  gboolean found; ///< Set to TRUE when the Component is found
};

/** Return the pointer to the Component in the Project tree store.

    \param iter is the iterator for the Component in the Project tree store.

*/
void
Component::componentProjectIter(GtkTreeIter *iter)
{
  GtkTreeStore *store;
  ComponentProjectIterContext context;

  context.iter = iter;
  context.name = getName();
  context.found = FALSE;

  switch (getType())
    {
    case QualResObject::DOCUMENT:
    case QualResObject::MEMO:
    case QualResObject::NARRATIVE_DOCUMENT:
      store = selectDocStore;
      break;
    case QualResObject::CATEGORY:
    case QualResObject::XREF:
    case QualResObject::FRAME:
      store = selectCatStore;
      break;
    case QualResObject::SET:
      store = selectSetStore;
      break;
    default:
      g_assert_not_reached();
      return;
    }
  gtk_tree_model_foreach(GTK_TREE_MODEL(store),
			 componentProjectIterForeachFunc,
			 (gpointer) &context);
  g_assert(context.found);
}

/** Function for componentProjectIter.
 */
gboolean
componentProjectIterForeachFunc(GtkTreeModel *model,
				GtkTreePath *path,
				GtkTreeIter *iter,
				gpointer data)
{
  ComponentProjectIterContext *context = (ComponentProjectIterContext *) data;
  gchar *gName;
  string name;

  // Get the Component name.
  gtk_tree_model_get(model,
		     iter,
		     QualResObject::FULL_NAME,
		     &gName,
		     -1);
  name = gName;
  g_free(gName);

  // Is it the component we want?
  if (name == context->name)
    {
      // Yes.  Stop looking.
      context->found = TRUE;
      *(context->iter) = *iter;
      return(TRUE);
    }

  // No.  Keep looking.
  return(FALSE);
}
