/*
 * $Id: Document.cc,v 1.19 2003/04/02 01:42:46 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Document.h"
#include "Category.h"
#include "qualresUtility.h"
#include "support.h"
#include <fstream>
#include <iostream>
#include <strstream>

/** Create empty Document.

    \param t is the Project tag table
 */
Document::Document(GtkTextTagTable *t)
{
  init(t);
}

/** Initialize the object.

    \param t is the Project tag table.

    This method creates a GtkTextBuffer.
*/
void
Document::init(GtkTextTagTable *t)
{
  objType = QualResObject::DOCUMENT;
  objName = "Document";
  xmlTag = DOCUMENT_TAG;
  projTagTable = t;
  textBuffer = gtk_text_buffer_new(projTagTable);
  // Update line numbers and position of Coded Text labels when text
  // is deleted.
  g_signal_connect(G_OBJECT(textBuffer),
		   "delete-range",
		   G_CALLBACK(textDeleted_cb),
		   NULL);
  // Update line numbers and position of Coded Text labels when text
  // is inserted.
  g_signal_connect(G_OBJECT(textBuffer),
		   "insert-text",
		   G_CALLBACK(textInserted_cb),
		   this);
  createWindow(NULL);
  setEditable(FALSE);
}

/** Destructor.
 */
Document::~Document()
{
  map<gint, CategoryLabels *>::iterator iCatLab;

  for (iCatLab = categoryLabels.begin();
       iCatLab != categoryLabels.end();
       iCatLab++)
    delete iCatLab->second;

}

/** Import a file into the buffer where the cursor is.

    \param fileName is the name of the file to import
*/
void
Document::importFile(string &fileName)
{
  ostrstream buf;
  string stringBuf;
  ifstream file;

  GtkWidget *dialog;
  string dialogMessage;

  file.open(fileName.c_str(), ios::in | ios::nocreate);
  if (file.bad())
    {
      dialogMessage = "Error opening file '" + fileName + "'";
      dialog = gtk_message_dialog_new(GTK_WINDOW(mainWindow),
				      GTK_DIALOG_DESTROY_WITH_PARENT,
				      GTK_MESSAGE_ERROR,
				      GTK_BUTTONS_CLOSE,
				      dialogMessage.c_str());
      gtk_dialog_run(GTK_DIALOG(dialog));
      gtk_widget_destroy(dialog);
      return;
    }
  buf << file.rdbuf() << ends;
  stringBuf = buf.str();
  buf.rdbuf()->freeze(0); // Unfreeze memory

  gtk_text_buffer_insert_at_cursor(textBuffer,
				   stringBuf.c_str(),
				   stringBuf.length());
}

/** Write to an XML element.

    \param xmlDoc is the XML document

    \return a pointer to the XML element

    The objects creates an XML element, and writes its persistent data
    to it.  For example, the buffer is saved, but the cursor positon
    is not.
*/
GdomeElement *
Document::writeXML(GdomeDocument *xmlDoc)
  throw(FileException)
{
  GdomeElement *elem;
  GdomeText *descNode;
  GdomeException exc;
  GdomeDOMString *tempStr, *tempVal;

  gchar *docText; // Document text
  GtkTextIter startIter, endIter; // Start/end iterators of textBuffer

  // Create element.
  tempStr = gdome_str_mkref(xmlTag);
  elem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
  gdome_str_unref(tempStr);
  if (elem == NULL)
    throw FileException();

  try
    {
      // Create and write the name element.
      setTextInElement(xmlDoc, elem,
		       NAME_TAG,
		       (gchar *) (getName()).c_str());

      // Create and write the description element.
      setTextInElement(xmlDoc, elem,
		       DESCRIPTION_TAG,
		       (gchar *) (getDesc()).c_str());
    }
  catch(FileException)
    {
      gdome_el_unref(elem, &exc);
      throw FileException();
    }

  // Get the text buffer.
  gtk_text_buffer_get_bounds(textBuffer,
			     &startIter,
			     &endIter);
  docText = gtk_text_buffer_get_slice(textBuffer,
				      &startIter,
				      &endIter,
				      TRUE);
  try
    {
      // Create and write the text element.
      setTextInElement(xmlDoc, elem, TEXT_TAG, docText);
    }
  catch(FileException)
    {
      g_free(docText);
      gdome_el_unref(elem, &exc);
      throw FileException();
    }

  g_free(docText);

  setSaved();
  return(elem);
}

/** Read from an XML element.

    \param elem is the XML element containing the object data
*/
bool
Document::readXML(GdomeElement *elem)
  throw(FileException)
{
  GtkTextIter startIter, endIter; // Start/end iterators of textBuffer

  // Load the name.
  pendingSetName(getTextFromElement(elem, NAME_TAG));
  commitName();

  // Load the description.
  setDesc(getTextFromElement(elem, DESCRIPTION_TAG));

  // Load the buffer.
  gtk_text_buffer_get_bounds(textBuffer,
			     &startIter,
			     &endIter);
  gtk_text_buffer_delete(textBuffer,
			 &startIter,
			 &endIter);
  gtk_text_buffer_insert(textBuffer,
			 &startIter,
			 (getTextFromElement(elem, TEXT_TAG)).c_str(),
			 -1);
  setSaved();
  return(TRUE);
}

/** Inform the object that a component is about to be deleted.

    \param comp is the component
*/
void
Document::pendingDelete(QualResObject *comp)
{
  map<gint, CategoryLabels *>::iterator iCatLab;

  // If this Document is to be Deleted, hide the Selection Windows.
  if ((Document *) comp == this)
    {
      hideAllSelectWindows();
      return;
    }

  // Unselect this Category if it was the destination Category for coding.
  if ((DocCat *) comp == selectedCat)
    setSelectedCategory("");

  // Remove all labels for this Category.
  if (comp->getType() == CATEGORY)
    {
      for (iCatLab = categoryLabels.begin();
	   iCatLab != categoryLabels.end();
	   iCatLab++)
	{
	  // Remove the Category label.
	  (iCatLab->second)->deleteCategoryLabel((Category *) comp);

	  // Remove the object if there are no more labels.
	  if ((iCatLab->second)->size() == 0)
	    {
	      delete iCatLab->second;
	      categoryLabels.erase(iCatLab->first);
	    }
	}
    }
}

/** Inform the object that a component is about to be cut.

    \param comp is the component
*/
void
Document::pendingCut(QualResObject *comp)
{
  // If this Document is to be Deleted, hide the Selection Windows.
  if ((Document *) comp == this)
    {
      hideAllSelectWindows();
      return;
    }

  // Unselect this Category if it was the destination Category for coding.
  if ((DocCat *) comp == selectedCat)
    setSelectedCategory("");
}

/** Make the Document buffer editable or non-editable.

    \param e is TRUE to set the buffer editable, false otherwise

 */
void
Document::setEditable(gboolean e)
{
  editable = e;
  gtk_text_view_set_editable(GTK_TEXT_VIEW(textView), editable);

  if (editable)
    gtk_label_set_label(GTK_LABEL(editableLabel),
			"Editable");
  else
    gtk_label_set_label(GTK_LABEL(editableLabel),
			"Not editable");
}

/** Get the number of lines to print.

    \returns the number of lines to print
*/
gint
Document::getPrintLineCount()
{
  return(gtk_text_buffer_get_line_count(textBuffer));
}

/** Get the text to print.

    \returns a pointer to allocated memory containing the text to print.
    This memory must be freed by g_free().
*/
gchar *
Document::getTextToPrint() const
{
  GtkTextIter start, end;

  gtk_text_buffer_get_bounds(textBuffer,
			     &start,
			     &end);
  
  return(gtk_text_buffer_get_text(textBuffer,
				   &start,
				   &end,
				   TRUE));
}

/** Test whether there is selected text in the Component.

    \returns TRUE if there's selected text in the Component,
    FALSE otherwise.
*/
gboolean
Document::hasSelectedText()
{
  return(gtk_text_buffer_get_selection_bounds(textBuffer, NULL, NULL));
}

/** Get the selected text to print.

    \returns a pointer to allocated memory containing the selected
    text to print.  This memory must be freed by g_free().
*/
gchar *
Document::getSelectedTextToPrint() const
{
  GtkTextIter start, end;

  gtk_text_buffer_get_selection_bounds(textBuffer,
				       &start,
				       &end);
  
  return(gtk_text_buffer_get_text(textBuffer,
				   &start,
				   &end,
				   TRUE));
}

/** Get the line number of the first line of the selected text to print.

    \returns the line number of the first line of the selected text to print.
*/
gint
Document::getSelectedTextToPrintLineNum()
{
  GtkTextIter start, end;

  gtk_text_buffer_get_selection_bounds(textBuffer,
				       &start,
				       &end);
  
  return(gtk_text_iter_get_line(&start) + 1);
}


/** Get the Component name labeled with "Document:  ".

    \returns left header of print output (free with g_free)
**/
gchar *
Document::getLabeledName() const
{
  return(g_strconcat("Document:  ", (getPrettyName()).c_str(), NULL));
}

/** Has the object been saved?
 */
bool
Document::saved() const
{
  return(SaveObject::saved() && !gtk_text_buffer_get_modified(textBuffer));
}

/** Specify that the object have been saved.
 */
void
Document::setSaved()
{
  SaveObject::setSaved();
  gtk_text_buffer_set_modified(textBuffer, FALSE);
}

/** Search for and code text into the selected Category.

    \param search is the string to search for
    \param searchType is the type of search
    \param cat is the selected DocCat
*/
void
Document::codeSearchText(string &search,
			 DocCat::SearchStore searchType,
			 Category *destCat)
{
  // Iterators.
  GtkTextIter startIter; // Start of search
  GtkTextIter matchStart, matchEnd; // Boundaries of found text

  gtk_text_buffer_get_start_iter(textBuffer, &startIter);

  // Code while a match is found.
  while (gtk_text_iter_forward_search(&startIter,
				      search.c_str(),
				      GTK_TEXT_SEARCH_TEXT_ONLY,
				      &matchStart, &matchEnd,
				      NULL))
    {
      // Next search begins with the end iterator.
      startIter = matchEnd;

      // Move the iterators, if requested.
      switch (searchType)
	{
	case DocCat::WORD:
	  // Move the iterators.
	  if (gtk_text_iter_inside_word(&matchStart))
	    gtk_text_iter_backward_word_start(&matchStart);
	  if (gtk_text_iter_inside_word(&matchEnd))
	    gtk_text_iter_forward_word_end(&matchEnd);
	  break;
	case DocCat::SENTENCE:
	  moveToSentenceBounds(&matchStart, &matchEnd, 0);
	  break;
	case DocCat::SENTENCE2:
	  moveToSentenceBounds(&matchStart, &matchEnd, 2);
	  break;
	case DocCat::STATEMENT:
	  moveToStatementBounds(&matchStart, &matchEnd);
	  break;
	case DocCat::BUFFER:
	  gtk_text_buffer_get_bounds(textBuffer, &matchStart, &matchEnd);
	  break;
	}

      // Code.
      destCat->codeDocText(this, &matchStart, &matchEnd, Category::UPDATE);
    }
}

/** Change the object's properties.

    \param parent is the parent window

 */
bool
Document::changeProperties(GtkWindow *parent)
{
  GtkWidget *table;
  GtkAttachOptions labelAttach = GTK_SHRINK, textAttach = GTK_FILL;
  GtkWidget *scrolledView, *view, *hbox;
  GtkTextIter start, end;
  bool status;
  gint result;

  GtkTreeSelection *selection;
  gchar *name;
  GtkTreeIter iter;

  string sProp, prompt;

  // Create the properties window.
  if (propWindow == NULL)
    {
      propWindow = gtk_dialog_new_with_buttons("Properties",
					       GTK_WINDOW(mainWindow),
					       GTK_DIALOG_MODAL,
					       GTK_STOCK_CANCEL,
					       GTK_RESPONSE_REJECT,
					       GTK_STOCK_OK,
					       GTK_RESPONSE_ACCEPT,
					       NULL);
      table = gtk_table_new(2, 3, TRUE);

      // Add label and text entry for the name.
      prompt = objName + " name:  ";
      gtk_table_attach(GTK_TABLE(table),
		       gtk_label_new(_(prompt.c_str())),
		       0, 1, 0, 1,
		       labelAttach, labelAttach,
		       5, 5);
      compName = gtk_entry_new();
      gtk_table_attach(GTK_TABLE(table),
		       compName,
		       1, 2, 0, 1,
		       textAttach, textAttach,
		       5, 5);

      // Add label and text buffer for the description.
      prompt = objName + " description:  ";
      gtk_table_attach(GTK_TABLE(table),
		       gtk_label_new(_(prompt.c_str())),
		       0, 1, 1, 2,
		       labelAttach, labelAttach,
		       5, 5);

      // Create view, and put it into a scrolled window.
      view = gtk_text_view_new();
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), view);

      compDesc = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
      gtk_table_attach(GTK_TABLE(table),
		       scrolledView,
		       1, 2, 1, 2,
		       textAttach, textAttach,
		       5, 5);

      // Create the editable radio buttons.
      gtk_table_attach(GTK_TABLE(table),
		       gtk_label_new(_("Editable buffer:  ")),
		       0, 1, 2, 3,
		       labelAttach, labelAttach,
		       5, 5);

      hbox = gtk_hbox_new(FALSE, 5);
      editableButtonY = gtk_radio_button_new_with_label(NULL, "Y");
      gtk_box_pack_start(GTK_BOX(hbox),
			 editableButtonY,
			 FALSE, FALSE, 0);
      editableButtonN =gtk_radio_button_new_with_label_from_widget
	(GTK_RADIO_BUTTON(editableButtonY), "N"),
      gtk_box_pack_start(GTK_BOX(hbox),
			 editableButtonN,
			 FALSE, FALSE, 0);

      gtk_table_attach(GTK_TABLE(table),
		       hbox,
		       1, 2, 2, 3,
		       textAttach, textAttach,
		       5, 5);

      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(propWindow)->vbox),
			table);
    }

  // Load the properties.
  sProp = getShortName();
  gtk_entry_set_text(GTK_ENTRY(compName), sProp.c_str());
  gtk_text_buffer_get_bounds(compDesc, &start, &end);
  gtk_text_buffer_delete(compDesc, &start, &end);
  sProp = getDesc();
  gtk_text_buffer_set_text(compDesc, sProp.c_str(), -1);
  if (getEditable())
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(editableButtonY),
				 TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(editableButtonN),
				 TRUE);

  // Show the properties window.
  gtk_widget_show_all(propWindow);
  result = gtk_dialog_run(GTK_DIALOG(propWindow));
  gtk_widget_hide(propWindow);

  // Load new properties.
  if (result == GTK_RESPONSE_ACCEPT)
    {
      sProp = gtk_entry_get_text(GTK_ENTRY(compName));
      sProp = getParentsName() + sProp;
      pendingSetName(sProp);
      gtk_text_buffer_get_bounds(compDesc, &start, &end);
      setDesc(gtk_text_buffer_get_text(compDesc, &start, &end, FALSE));
      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(editableButtonY)))
#if 0
	setEditable(TRUE);
#endif
      showMessage("Editing is disabled in this version of qualres",
		  GTK_MESSAGE_INFO,
		  mainWindow);

      else
	setEditable(FALSE);
      status = true;
    }
  else
    status = false;

  return(status);
}

/** Create the window to code and view a Document.

    \param parent is the parent window

 */
void
Document::createWindow(GtkWidget *parent)
{
  /* Menus */
  static GnomeUIInfo fileMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Import from file"),
     N_("Import text from a file to the cursor position"),
     (gpointer) documentImport_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CONVERT,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Export Document"),
     N_("Export the Document to a file"),
     (gpointer) docCatExport_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CONVERT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Print preview"),
     N_("Print preview"),
     (gpointer) printPreview_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT_PREVIEW,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Print"),
     N_("Print the Document"),
     (gpointer) print_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Close window"),
     N_("Close the Document window"),
     (gpointer) close_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CLOSE,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo catMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Select Category"),
     N_("Select the Category for Coding Text"),
     (gpointer) selectCodeCategoryList_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Open Selected Category"),
     N_("Open the Selected Category"),
     (gpointer) codeViewSelectedCat_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_OPEN,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo help_menu [] = {
    GNOMEUIINFO_HELP ("Document"),
  
    GNOMEUIINFO_END
  };

  static GnomeUIInfo menu [] = {
    { GNOME_APP_UI_SUBTREE_STOCK, N_("File"), NULL, fileMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    { GNOME_APP_UI_SUBTREE_STOCK, N_("Category"), NULL, catMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    GNOMEUIINFO_MENU_HELP_TREE(help_menu),
    GNOMEUIINFO_END
  };

  GtkWidget *status, *vbox, *hbox, *frame;
  GtkWidget *scrolledView, *view;
  GtkWidget *button;
  GtkAdjustment *adj;
  
  // Create the main window.
  if (mainWindow == NULL)
    {
      // Create widgets.
      mainWindow = gnome_app_new("Document", _("Document a"));
      gtk_window_set_default_size(GTK_WINDOW(mainWindow),
				  DOCUMENT_XSIZE,
				  DOCUMENT_YSIZE);

      status = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_NEVER);

      textView = gtk_text_view_new_with_buffer(textBuffer);
      gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(textView),
				  GTK_WRAP_WORD);
      gtk_text_view_set_border_window_size(GTK_TEXT_VIEW(textView),
					   GTK_TEXT_WINDOW_LEFT,
					   1);
      g_signal_connect(G_OBJECT(textView),
		       "expose-event",
		       G_CALLBACK(text_view_line_numbers_expose_cb),
		       textBuffer);

      // Update placement of Coded Text labels when the size of the
      // text view changes.
      g_signal_connect(G_OBJECT(textView),
		       "size-allocate",
		       G_CALLBACK(textSizeChanged_cb),
		       this);

      frame = gtk_frame_new(NULL);
      gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
      gtk_signal_connect(GTK_OBJECT(mainWindow), "delete-event",
			 GTK_SIGNAL_FUNC(hide_cb), this);

      vbox = gtk_vbox_new(FALSE, 5);

      editableLabel = gtk_label_new(_("Editable"));
      gtk_box_pack_start(GTK_BOX(vbox),
			 editableLabel,
			 FALSE, FALSE, 0);

      hbox = gtk_hbox_new(FALSE, 5);

      button = gtk_button_new_with_label("Select");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(selectCodeCategoryList_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      gtk_box_pack_start(GTK_BOX(hbox),
			 gtk_label_new(_("Selected Category:  ")),
			 FALSE, FALSE, 0);

      selectedCatBuffer = gtk_text_buffer_new
	(gtk_text_buffer_get_tag_table(textBuffer));
      view = gtk_text_view_new_with_buffer(selectedCatBuffer);
      g_signal_connect(G_OBJECT(view),
		       "button-press-event",
		       G_CALLBACK(selectedCatViewButtonPress_cb),
		       this);
      gtk_text_view_set_editable(GTK_TEXT_VIEW(view),
				 FALSE);
      gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view),
				       FALSE);

      gtk_box_pack_start(GTK_BOX(hbox),
			 view,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Code");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(documentCode_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Uncode");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(documentUncode_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox),
			 hbox,
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_hseparator_new(),
			 FALSE, FALSE, 0);

      GtkWidget *hPane;
      GdkColor color;

      color.red = 0xFFFF;
      color.green = 0xFFFF;
      color.blue = 0xFFFF;

      // Widget for Category labels.
      layout = gtk_layout_new(NULL, NULL);
      gtk_widget_modify_bg(layout,
			   GTK_STATE_NORMAL,
			   &color);
      gtk_layout_set_size(GTK_LAYOUT(layout), 2000, 100);

      // Horizontal pane.
      hPane = gtk_hpaned_new();

      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      adj = gtk_scrolled_window_get_vadjustment
	(GTK_SCROLLED_WINDOW(scrolledView));

      // Update placement of Coded Text labels when the text view is scrolled.
      g_signal_connect_after(G_OBJECT(adj),
			     "value-changed",
			     G_CALLBACK(textScrolled_cb),
			     this);

      gtk_container_add(GTK_CONTAINER(scrolledView),
			textView);
      gtk_paned_pack1(GTK_PANED(hPane), scrolledView, FALSE, TRUE);

      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView),
			layout);
      gtk_paned_pack2(GTK_PANED(hPane), scrolledView, TRUE, TRUE);
      gtk_paned_set_position(GTK_PANED(hPane),
			     DOCUMENT_XSIZE - DOCUMENT_XSIZE / 5);

      gtk_box_pack_start(GTK_BOX(vbox),
			 hPane,
			 TRUE, TRUE, 0);

      // Add to application window.
      gnome_app_set_contents(GNOME_APP(mainWindow), frame);
      gnome_app_set_statusbar(GNOME_APP(mainWindow), status);
      gnome_app_create_menus_with_data(GNOME_APP(mainWindow), menu, this);
      gnome_app_install_menu_hints(GNOME_APP(mainWindow), menu);

      // Add the widgets.
      gtk_container_add(GTK_CONTAINER(frame), vbox);

    }
}

/** Scroll the buffer so that a specified mark is visible.

    \param mark is the mark to make visible

    The buffer is scrolled until the mark is visible.  Exactly where
    the mark is positioned is controlled by the parameters
    DOC_MARK_VISIBLE_X and DOC_MARK_VISIBLE_Y.

*/
void
Document::makeVisible(GtkTextMark *mark)
{
  GtkTextIter iter;
  GtkTextMark *tempMark;

  // Move the cursor to the mark.
  gtk_text_buffer_get_iter_at_mark(textBuffer,
				   &iter,
				   mark);
  gtk_text_buffer_place_cursor(textBuffer, &iter);

  // Find the mark at the start of the line.
  gtk_text_iter_set_line_offset(&iter, 0);

  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(textView),
			       &iter,
			       0.0,
			       TRUE,
			       DOC_MARK_VISIBLE_X,
			       DOC_MARK_VISIBLE_Y);

  tempMark = gtk_text_buffer_create_mark(textBuffer,
					 NULL,
					 &iter,
					 FALSE);

  gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(textView),
				     tempMark);

  // Move so that the mark is visible.
  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textView),
			       tempMark,
			       0.0,
			       TRUE,
			       DOC_MARK_VISIBLE_X,
			       DOC_MARK_VISIBLE_Y);
  gtk_text_buffer_delete_mark(textBuffer,
			      tempMark);
}

/** Add a Category label, if there isn't one.

    \param cat is the Category the label is for

    \param textIter is the location to add it

*/
void
Document::addCategoryLabel(Category *cat, GtkTextIter *textIter)
{
  gint lineNum;
  gint yPixel, yHeight;
  gint windowX, windowY;

  g_assert(cat != NULL);

  // Get the line number where the Category label is to be placed.
  lineNum = gtk_text_iter_get_line(textIter);

  // Create the object to hold all the Category labels for this line.
  if (categoryLabels.count(lineNum) == 0)
      categoryLabels[lineNum] = new CategoryLabels;

  // Add the Category.
  categoryLabels[lineNum]->createCategoryLabel(cat);

  // If the labels for this line have not yet been put into the label
  // widget, put them there.
  if ( ! (categoryLabels[lineNum]->getPutStatus()) )
    {
      // Calculate the coordinates in the label widget.
      gtk_text_view_get_line_yrange(GTK_TEXT_VIEW(textView),
				    textIter,
				    &yPixel,
				    &yHeight);

      gtk_text_view_buffer_to_window_coords(GTK_TEXT_VIEW(textView),
					    GTK_TEXT_WINDOW_TEXT,
					    0,
					    yPixel,
					    &windowX,
					    &windowY);

      // Put the widget into the label.
      gtk_layout_put(GTK_LAYOUT(layout),
		     categoryLabels[lineNum]->getHBox(),
		     0,
		     windowY);
  
      gtk_widget_show_all(layout);

      // Store the fact that the widget was put.
      categoryLabels[lineNum]->setPutStatus(TRUE);
    }
}

/** Remove a Category label, if appropriate.

    \param cat is the Category the label is for

    \param textIter is the location of the Category label to remove

    This method check how many Coded Text segments begin at the
    specified location.  If there are none, the label is removed.
    Otherwise, the label is not removed.

    If there are no labels, the method does nothing.

*/
void
Document::removeCategoryLabel(Category *cat, GtkTextIter *textIter)
{
  gint lineNum;

  g_assert(cat != NULL);
  g_assert(textIter != NULL);

  // Get the line number where the Category label is to be removed.
  lineNum = gtk_text_iter_get_line(textIter);

  // Return if there are no labels.
  if (categoryLabels.count(lineNum) == 0)
    return;

  // Check if this Category has any Coded Text beginning on this line.
  if (cat->hasLabels(this, lineNum))
    return; // Yes.  Do nothing.

  // The Category has no Coded Text beginning on this line.  Remove
  // its label.
  categoryLabels[lineNum]->deleteCategoryLabel(cat);

  // Remove the object if there are no more labels.
  if (categoryLabels[lineNum]->size() == 0)
    {
      delete categoryLabels[lineNum];
      categoryLabels.erase(lineNum);
    }
}

/** Adjust the positions of the existing Category labels.

    Each CategoryLabels object is repositioned.

*/
void
Document::adjustCategoryLabels()
{
  GtkTextIter iter;
  gint lineNum, yPixel, yHeight, windowX, windowY;
  map<gint, CategoryLabels *>::iterator iCatLabs;

  guint w, h;
  gtk_layout_get_size(GTK_LAYOUT(layout), &w, &h);

  for (iCatLabs = categoryLabels.begin();
       iCatLabs != categoryLabels.end();
       iCatLabs++)
    {
      lineNum = iCatLabs->first;

      // Get the text iterator for this line.
      gtk_text_buffer_get_iter_at_line(textBuffer,
				       &iter,
                                       lineNum);

      // Calculate the coordinates in the label widget.
      gtk_text_view_get_line_yrange(GTK_TEXT_VIEW(textView),
				    &iter,
				    &yPixel,
				    &yHeight);

      gtk_text_view_buffer_to_window_coords(GTK_TEXT_VIEW(textView),
					    GTK_TEXT_WINDOW_TEXT,
					    0,
					    yPixel,
					    &windowX,
					    &windowY);

      // Put the widget into the label widget.
      gtk_layout_move(GTK_LAYOUT(layout),
		      categoryLabels[lineNum]->getHBox(),
		      0,
		      windowY);
    }
}

/** Code or Uncode the selected text.

    \param codeUncode specifies whether to code or uncode

    This method obtains the iterators of the selected text, and passes
    them to the Category method for coding.  The selected text is
    unselected, the the focus is left in the text view widget.

    If there is no selection, or if there is no selected Category,
    this function does nothing.
*/
void
Document::codeUncodeText(enum CodeUncode codeUncode)
{
  GtkTextIter start, end, cursor;

  if (selectedCat == NULL)
    return;

  // Move the focus to the view.
  gtk_widget_grab_focus(textView);

  // Obtain the Iterators.
  if (!gtk_text_buffer_get_selection_bounds(textBuffer, &start, &end))
    return;

  // Unselect.
  gtk_text_buffer_get_iter_at_mark(textBuffer, 
				   &cursor,
				   gtk_text_buffer_get_insert(textBuffer));
  gtk_text_buffer_place_cursor(textBuffer, &cursor);

  // Code.
  if (codeUncode == DocCat::CODE)
    selectedCat->codeDocText(this,
			     &start,
			     &end,
			     Category::UPDATE);
  else
    selectedCat->uncodeDocText(this,
			       &start,
			       &end);
}

/***********************************************************************/
/* CALLBACKS */
/***********************************************************************/
/** Code text in a Document.
*/
void
documentCode_cb(GtkWidget *widget, gpointer data)
{
  Document *doc = (Document *) data;

  doc->codeUncodeText(DocCat::CODE);
}

/** Uncode text in a Document.
*/
void
documentUncode_cb(GtkWidget *widget, gpointer data)
{
  Document *doc = (Document *) data;

  doc->codeUncodeText(DocCat::UNCODE);
}

/** Import text to a Document.
*/
void
documentImport_cb(GtkWidget *widget, gpointer data)
{
  Document *doc = (Document *) data;
  GtkWidget *fileSelect;
  string fileName;
  gint result;

  // Pop up file selection box.
   fileSelect = gtk_file_selection_new("Please select a file for import");
   
   // Display.
   gtk_widget_show(fileSelect);
   result = gtk_dialog_run(GTK_DIALOG(fileSelect));
   fileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fileSelect));
   gtk_widget_destroy(GTK_WIDGET(fileSelect));

  // Import file.
  if (result == GTK_RESPONSE_OK)
    {
      if (fileName.substr(fileName.length() - 1, 1) != "/")
	doc->importFile(fileName);
    }
}

/** Adjust the Category view widgets when text is deleted from a Document.
 */
void
textDeleted_cb(GtkTextBuffer *textBuffer,
	       GtkTextIter *startIter,
	       GtkTextIter *endIter,
	       gpointer userData)
{
  gchar *docText;

#if 0
  docText = gtk_text_buffer_get_text(textBuffer,
				     startIter,
				     endIter,
				     TRUE);
  cerr << "Deleted text '" << docText << "'\n";
  g_free(docText);
#endif
}

/** Adjust the Category view widgets when text is added to a Document.
 */
void
textInserted_cb(GtkTextBuffer *textBuffer,
		GtkTextIter *startIter,
		gchar *text,
		gint nChars,
		gpointer data)
{
  Document *doc = (Document *) data;

#if 0
  cerr << nChars << " characters inserted at line "
       << gtk_text_iter_get_line(startIter) << endl;
#endif
}

/** Adjust the Category view widgets when the view size is changed.
 */
void
textSizeChanged_cb(GtkWidget *widget,
		   GtkAllocation *allocation,
		   gpointer data)
{
  Document *doc = (Document *) data;

  doc->adjustCategoryLabels();
}

/** Adjust the Category view widgets when the view is scrolled.
 */
void
textScrolled_cb(GtkAdjustment *adj,
		gpointer data)
{
  Document *doc = (Document *) data;

  doc->adjustCategoryLabels();
}
