/*
 * $Id: Document.h,v 1.7 2003/04/01 11:25:34 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifndef DOCUMENT_H
#define DOCUMENT_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string>
#include "qualres.h"
#include "DocCat.h"
#include "CategoryLabels.h"


/** Class for a Document

    This class implements the methods for a Document.

    All access functions of this and derived classes return a pointer
    to const.  For example, getName returns (const string *), pointing
    to the name field.  The object is responsible for this memory; the
    calling program should never free the memory pointed to by this
    pointer.
*/
class Document : public DocCat
{
public:
  Document() {} ///< Constructor for derived classes
  Document(GtkTextTagTable *);
  virtual ~Document();

  virtual bool saved() const;

  virtual void pendingDelete(QualResObject *);
  virtual void pendingCut(QualResObject *);

  void setEditable(gboolean);
  gboolean getEditable() const { return(editable); } ///< Returns editability.

 // Printing methods.
  virtual gint getPrintLineCount();
  virtual gchar *getTextToPrint() const;
  virtual gboolean hasSelectedText();
  virtual gchar *getSelectedTextToPrint() const;
  virtual gint getSelectedTextToPrintLineNum();
  virtual gint printLineNumbers()   ///< How often to number lines.
    { return(1); }
  virtual gchar *getLabeledName() const;

  /** Get the Documents GtkTextBuffer.

      \return GtkTextBuffer of the Document
  */
  GtkTextBuffer *getTextBuffer() const { return(textBuffer); }

  /** Get the Documents GtkTextView.

      \return GtkTextView of the Document
  */
  GtkWidget *getTextView() const { return(textView); }

  void importFile(string &);

  virtual GdomeElement *writeXML(GdomeDocument *)
    throw(FileException);
  virtual bool readXML(GdomeElement *)
    throw(FileException);

  virtual void codeUncodeText(enum CodeUncode);

  /// Search for and code text into the selected Category.
  virtual void codeSearchText(string &, DocCat::SearchStore, Category *);

  // GUI methods.
  virtual bool changeProperties(GtkWindow *);
  virtual void createWindow(GtkWidget *);
  virtual void makeVisible(GtkTextMark *);
  virtual void addCategoryLabel(Category *, GtkTextIter *);
  virtual void removeCategoryLabel(Category *, GtkTextIter *);
  virtual void adjustCategoryLabels();
  virtual void setSaved();

protected:
  GtkTextBuffer *textBuffer; ///< Text buffer.
  GtkWidget *textView; ///< View of text buffer.
  GtkWidget *layout; ///< Widget for Category labels.
  GtkWidget *editableLabel; ///< Editable label on main window.

private:
  void init(GtkTextTagTable *);
  gboolean editable; ///< TRUE if the buffer can be edited.
  GtkWidget *editableButtonY; ///< Radio button for editable.
  GtkWidget *editableButtonN; ///< Radio button for non-editable.
  map<gint, CategoryLabels *> categoryLabels; ///< Category labels.
};

// Callbacks.
void documentCode_cb(GtkWidget *, gpointer);
void documentUncode_cb(GtkWidget *, gpointer);
void documentImport_cb(GtkWidget *, gpointer);
void textDeleted_cb(GtkTextBuffer *, GtkTextIter *, GtkTextIter *, gpointer);
void textInserted_cb(GtkTextBuffer *, GtkTextIter *, gchar *, gint, gpointer);
void textSizeChanged_cb(GtkWidget *, GtkAllocation *, gpointer);
void textScrolled_cb(GtkAdjustment *, gpointer);

#endif
