/*
 * $Id: NarrativeDocument.cc,v 1.5 2003/04/02 01:42:46 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "NarrativeDocument.h"
#include "Frame.h"
#include "Category.h"
#include "qualresUtility.h"
#include "support.h"

#include <iostream>

/** Create empty NarrativeDocument.

    \param t is the Project tag table
 */
NarrativeDocument::NarrativeDocument(GtkTextTagTable *t)
{
  init(t);
}

/** Initialize the object.

    \param t is the Project tag table.

    This method creates a GtkTextBuffer.
*/
void
NarrativeDocument::init(GtkTextTagTable *t)
{
  objType = QualResObject::NARRATIVE_DOCUMENT;
  objName = "Narrative Document";
  xmlTag = NARRATIVE_DOCUMENT_TAG;
  projTagTable = t;
  textBuffer = gtk_text_buffer_new(projTagTable);
  // Update line numbers and position of Coded Text labels when text
  // is deleted.
  g_signal_connect(G_OBJECT(textBuffer),
		   "delete-range",
		   G_CALLBACK(textDeleted_cb),
		   NULL);
  // Update line numbers and position of Coded Text labels when text
  // is inserted.
  g_signal_connect(G_OBJECT(textBuffer),
		   "insert-text",
		   G_CALLBACK(textInserted_cb),
		   this);
  createWindow(NULL);
  setEditable(FALSE);
}

/** Destructor.
 */
NarrativeDocument::~NarrativeDocument()
{
}

/** Import a Document into the buffer where the cursor is.

*/
void
NarrativeDocument::importDocument()
{
  cerr << "NarrativeDocument::importDocument\n";
}

/** Get the Component name labeled with "NarrativeDocument:  ".

    \returns left header of print output (free with g_free)
**/
gchar *
NarrativeDocument::getLabeledName() const
{
  return(g_strconcat("Narrative Document:  ",
		     (getPrettyName()).c_str(), NULL));
}

/** Create the window to code and view a Narrative Analysis Document.

    \param parent is the parent window

 */
void
NarrativeDocument::createWindow(GtkWidget *parent)
{
  /* Menus */
  static GnomeUIInfo fileMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Import from file"),
     N_("Import text from a file to the cursor position"),
     (gpointer) documentImport_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CONVERT,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Export Narrative Document"),
     N_("Export the Narrative Document to a file"),
     (gpointer) docCatExport_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CONVERT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Print preview"),
     N_("Print preview"),
     (gpointer) printPreview_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT_PREVIEW,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Print"),
     N_("Print the Narrative Document"),
     (gpointer) print_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Close window"),
     N_("Close the Narrative Document window"),
     (gpointer) close_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CLOSE,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo frameMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Select Frame"),
     N_("Select the Frame for Coding Text"),
     (gpointer) selectCodeCategoryList_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo help_menu [] = {
    GNOMEUIINFO_HELP ("Narrative Analysis Document"),
  
    GNOMEUIINFO_END
  };

  static GnomeUIInfo menu [] = {
    { GNOME_APP_UI_SUBTREE_STOCK, N_("File"), NULL, fileMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    { GNOME_APP_UI_SUBTREE_STOCK, N_("Frame"), NULL, frameMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    GNOMEUIINFO_MENU_HELP_TREE(help_menu),
    GNOMEUIINFO_END
  };

  GtkWidget *status, *vbox, *hbox, *frame;
  GtkWidget *scrolledView, *view;
  GtkWidget *button;
  GtkAdjustment *adj;
  
  // Create the main window.
  if (mainWindow == NULL)
    {
      // Create widgets.
      mainWindow = gnome_app_new("Narrative Document", _("Document a"));
      gtk_window_set_default_size(GTK_WINDOW(mainWindow),
				  DOCUMENT_XSIZE,
				  DOCUMENT_YSIZE);

      status = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_NEVER);

      textView = gtk_text_view_new_with_buffer(textBuffer);
      gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(textView),
				  GTK_WRAP_WORD);
      gtk_text_view_set_border_window_size(GTK_TEXT_VIEW(textView),
					   GTK_TEXT_WINDOW_LEFT,
					   1);
      g_signal_connect(G_OBJECT(textView),
		       "expose-event",
		       G_CALLBACK(text_view_line_numbers_expose_cb),
		       textBuffer);

      // Update placement of Coded Text labels when the size of the
      // text view changes.
      g_signal_connect(G_OBJECT(textView),
		       "size-allocate",
		       G_CALLBACK(textSizeChanged_cb),
		       this);

      frame = gtk_frame_new(NULL);
      gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
      gtk_signal_connect(GTK_OBJECT(mainWindow), "delete-event",
			 GTK_SIGNAL_FUNC(hide_cb), this);

      vbox = gtk_vbox_new(FALSE, 5);

      editableLabel = gtk_label_new(_("Editable"));
      gtk_box_pack_start(GTK_BOX(vbox),
			 editableLabel,
			 FALSE, FALSE, 0);

      hbox = gtk_hbox_new(FALSE, 5);

      button = gtk_button_new_with_label("Select");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(selectCodeCategoryList_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      gtk_box_pack_start(GTK_BOX(hbox),
			 gtk_label_new(_("Selected Frame:  ")),
			 FALSE, FALSE, 0);

      selectedCatBuffer = gtk_text_buffer_new
	(gtk_text_buffer_get_tag_table(textBuffer));
      view = gtk_text_view_new_with_buffer(selectedCatBuffer);
      g_signal_connect(G_OBJECT(view),
		       "button-press-event",
		       G_CALLBACK(selectedCatViewButtonPress_cb),
		       this);
      gtk_text_view_set_editable(GTK_TEXT_VIEW(view),
				 FALSE);
      gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view),
				       FALSE);

      gtk_box_pack_start(GTK_BOX(hbox),
			 view,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Frame");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(documentCode_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Unframe");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(unframe_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox),
			 hbox,
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_hseparator_new(),
			 FALSE, FALSE, 0);

      GtkWidget *hPane;
      GdkColor color;

      color.red = 0xFFFF;
      color.green = 0xFFFF;
      color.blue = 0xFFFF;

      // Widget for Frame, Stanza, etc. labels.
      layout = gtk_layout_new(NULL, NULL);
      gtk_widget_modify_bg(layout,
			   GTK_STATE_NORMAL,
			   &color);
      gtk_layout_set_size(GTK_LAYOUT(layout), 2000, 100);

      // Horizontal pane.
      hPane = gtk_hpaned_new();

      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      adj = gtk_scrolled_window_get_vadjustment
	(GTK_SCROLLED_WINDOW(scrolledView));

      // Update placement of Coded Text labels when the text view is scrolled.
      g_signal_connect_after(G_OBJECT(adj),
			     "value-changed",
			     G_CALLBACK(textScrolled_cb),
			     this);

      gtk_container_add(GTK_CONTAINER(scrolledView),
			textView);
      gtk_paned_pack1(GTK_PANED(hPane), scrolledView, FALSE, TRUE);

      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView),
			layout);
      gtk_paned_pack2(GTK_PANED(hPane), scrolledView, TRUE, TRUE);
      gtk_paned_set_position(GTK_PANED(hPane),
			     DOCUMENT_XSIZE - DOCUMENT_XSIZE / 5);

      gtk_box_pack_start(GTK_BOX(vbox),
			 hPane,
			 TRUE, TRUE, 0);

      // Add to application window.
      gnome_app_set_contents(GNOME_APP(mainWindow), frame);
      gnome_app_set_statusbar(GNOME_APP(mainWindow), status);
      gnome_app_create_menus_with_data(GNOME_APP(mainWindow), menu, this);
      gnome_app_install_menu_hints(GNOME_APP(mainWindow), menu);

      // Add the widgets.
      gtk_container_add(GTK_CONTAINER(frame), vbox);

    }
}

/** Unframe text.

*/
void
NarrativeDocument::unframe()
{
  GtkTextIter textIter;
  GtkTextMark *textMark;
  map<const string, Category *>::iterator iFrame;

  // Obtain the iterator where the cursor is.
  textMark = gtk_text_buffer_get_insert(textBuffer);
  gtk_text_buffer_get_iter_at_mark(textBuffer,
				   &textIter,
				   textMark);

  // Request each Frame to Unframe the text.
  for (iFrame = projectCatMap->begin();
       iFrame != projectCatMap->end();
       iFrame++)
    {
      ((Frame *) iFrame->second)->unframeNDocText(this,
						  &textIter);
    }
}

/***********************************************************************/
/* CALLBACKS */
/***********************************************************************/

/** Import text from a Document.
*/
void
narrativeDocumentDocumentImport_cb(GtkWidget *widget, gpointer data)
{
  NarrativeDocument *nDoc = (NarrativeDocument *) data;

  nDoc->importDocument();
}

/** Unframe text in a Narrative Document.
*/
void
unframe_cb(GtkWidget *widget, gpointer data)
{
  NarrativeDocument *nDoc = (NarrativeDocument *) data;

  nDoc->unframe();
}
