/*
 * $Id: Project.h,v 1.15 2003/04/16 14:11:47 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifndef PROJECT_H
#define PROJECT_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string>
#include <libgdome/gdome.h>

#include "qualres.h"
#include "Document.h"
#include "Category.h"
#include "Set.h"
#include "Memo.h"
#include "Xref.h"
#include "NarrativeDocument.h"
#include "Frame.h"

/** Class for a Project

    This class implements the methods for a Project, and contains all
    the Project Components.  The Project object is responsible for
    creating and destroying the Project Components.

    All access functions of this and derived classes return a pointer
    to const.  For example, getName returns (const string *), pointing
    to the name field.  The object is responsible for this memory; the
    calling program should never free the memory pointed to by this
    pointer.
*/
class Project: public QualResObject
{
public:
  /** Flag to determine whether the operation is Cut or Delete.
   */
  enum CutDelete
  {
    CUT = 0,
    DELETE,
    NCUT_DELETE
  };

  /** Status to indicate whether a Project is to be saved or discarded.
   */
  enum SaveDiscard
  {
    SAVE = 1,
    DISCARD,
    CANCEL = GTK_RESPONSE_CANCEL,
    NSAVE_DISCARD
  };

  /** Status to indicate whether to autorecover.
   */
  enum AutoRecover
  {
    READ_AUTOSAVE = 1,
    READ_PROJECT,
    CANCEL_READ,
    NAUTO_RECOVER
  };

  Project(GtkWidget *);
  virtual ~Project();

  void setFileName(string);

  /** Get the object's disk file name.

      \return the disk file name
  */
  const string &getFileName() const { return(fileName); }

  void createComponent(QualResObject::Type, QualResObject::Relationship);
  void updateComponent(QualResObject::Type);
  void removeComponent(QualResObject::Type, Project::CutDelete);
  void pasteComponent(QualResObject::Type, QualResObject::Relationship);
  void changeComponentProperties(QualResObject::Type);

  string getComponentList(QualResObject::Type);
  void printComponentList(QualResObject::Type);
  void exportComponentList(QualResObject::Type);

  void printAllComponents(QualResObject::Type);

  virtual void pendingCut(QualResObject *);
  virtual void pendingPaste(QualResObject *);
  virtual void pendingDelete(QualResObject *);
  virtual void propertiesHaveChanged(QualResObject *);

  void clearComponents();

  bool projectSaved();
  enum SaveDiscard  saveToFile();
  void saveAsToFile();
  void restoreFromFile();
  void restore();
  enum SaveDiscard closeProject();

  virtual GdomeElement *writeXML(GdomeDocument *);
  virtual bool readXML(GdomeElement *);

  void search();

  void setMergeCatDest(string);
  void getAndSetMergeCatDest();
  void mergeCatMerge();

  gboolean timeout();
  gboolean idle();

  // GUI methods.
  virtual bool changeProperties(GtkWindow *);
  virtual void createWindow(GtkWidget *);
  void updateTreeStore(QualResObject::Type);
  void createSearchWindow(GtkWidget *);
  void showSearchWindow();
  void hideSearchWindow();
  void selectSearchCategoryList();
  void loadSearchCategoryFromList();
  void openSelectedComponent(GtkWidget *);
  void popupComponentMenu(GtkWidget *, GdkEventButton *);
  void createMemoXrefWindow(GtkWidget *);
  void showMemoXrefWindow();
  void hideMemoXrefWindow();
  void createMergeCatWindow(GtkWidget *);
  void showMergeCatWindow();
  void createNarrativeWindow(GtkWidget *);
  void showNarrativeWindow();
  void hideNarrativeWindow();

private:
  void save();
  void autoSave();
  enum AutoRecover autoRecoverFile(string &, string &, GtkWidget *);
  void saveToRoot(GdomeDocument *, GdomeElement *);
  void close();
  void restoreComponents(GdomeNodeList *, QualResObject::Type)
    throw(FileException);
  void initComponent(Component *);
  void updateChildrenName(map<const string, Component *> *,
			  GtkTreeStore *,
			  GtkTreeIter *,
			  string &,
			  string &);

  vector<Document *> cutDocs; ///< Container of cut Documents.
  vector<Category *> cutCats; ///< Container of cut Categories.
  vector<Set *> cutSets; ///< Container of cut Sets.
  vector<Memo *> cutMemos; ///< Container of cut Memos.
  vector<Xref *> cutXrefs; ///< Container of cut Xrefs.

  /// Container of cut Narrative Analysis Documents.
  vector<NarrativeDocument *> cutNDocs;

  /// Container of cut Frames.
  vector<Frame *> cutFrames;

  /// Container of Document pointers
  map<const string, Document *> docMap;

  /// Container of Category pointers
  map<const string, Category *> catMap;

  /// Container of Set pointers
  map<const string, Set *> setMap;

  /// Container of Memo pointers
  map<const string, Memo *> memoMap;

  /// Container of Cross-reference pointers
  map<const string, Xref *> xrefMap;

  /// Container of Narrative Analysis Document pointers
  map<const string, NarrativeDocument *> narrativeMap;

  /// Container of Frame pointers
  map<const string, Frame *> frameMap;

  /// Empty container.
  map<const string, Set *> emptyMap;

  string fileName; ///< Name of disk file storing this Project
  string autoSaveName; ///< Name of autosave file storing this Project

  // Widgets.
  GtkWidget *docView; ///< View widget for Document tree
  GtkWidget *catView; ///< View widget for Category tree
  GtkWidget *setView; ///< View widget for Set tree
  GtkTreeStore *docStore; ///< Document tree model
  GtkTreeStore *catStore; ///< Category tree model
  GtkTreeStore *setStore; ///< Set tree model

  GtkTextTagTable *tagTable; ///< Tag table for Categories

  // Search window widgets.
  GtkWidget *searchWindow; ///< Search window
  GtkWidget *searchDocView; ///< View widget for Documents to search
  GtkWidget *searchCatView; ///< View widget for Categories to search
  GtkWidget *searchSetView; ///< View widget for Sets to search
  GtkWidget *searchText; ///< Search string
  GtkTextBuffer *searchCatBuffer; ///< Search destination Category name
  GtkWidget *searchTypeOptionMenu; ///< Search type (word, sentence, etc)

  Category *searchDestCat; ///< Search destination Category

  GtkWidget *selectSearchCatView; ///< Project Category tree view
  GtkWidget *selectSearchCatWindow; ///< Popup window for selecting a Category

  // Popup menu widgets.
  GtkWidget *popupDocMenu; ///< Popup Document menu.
  GtkWidget *popupCatMenu; ///< Popup Category menu.
  GtkWidget *popupSetMenu; ///< Popup Set menu.

  GtkWidget *popupMemoMenu; ///< Popup Memo menu.
  GtkWidget *popupXrefMenu; ///< Popup Cross-reference menu.

  GtkWidget *popupNarrativeMenu; ///< Popup Narrative Analysis menu.
  GtkWidget *popupFrameMenu; ///< Popup Frame menu.

  // Memo window widgets.
  GtkWidget *memoXrefWindow; ///< Memo window
  GtkWidget *memoView; ///< View widget for Memo tree
  GtkTreeStore *memoStore; ///< Memo tree model
  GtkWidget *xrefView; ///< View widget for Cross-reference tree
  GtkTreeStore *xrefStore; ///< Cross-reference tree model

  // Narrative Analysis widgets.
  GtkWidget *narrativeWindow; ///< Narrative Analysis window
  GtkWidget *narrativeView; ///< View widget for Narrative Documents tree
  GtkTreeStore *narrativeStore; ///< Narrative Documents tree model
  GtkWidget *frameView; ///< View widget for Frame tree
  GtkTreeStore *frameStore; ///< Frame tree model

  // Document filtering window widgets.
  GtkWidget *docFilterWindow; ///< Document filtering window
  GtkWidget *filteredMessage; ///< Widget containing the filtering message
  GtkTextBuffer *filteredBuffer; ///< Text of filtering message
  Set *selectedSet; ///< Selected Set for filtering (NULL if no filtering)

  // Widgets for merging Categories.
  GtkWidget *mergeCatWindow; ///< Toplevel window for merging Categories
  /// View widget to choose the Categories to merge.
  GtkWidget *mergeCatMergeView;
  /// View widget to choose the destination Category.
  GtkWidget *mergeCatDestView;
  /// Text buffer for the merge destination Category name.
  GtkTextBuffer *mergeCatDestBuffer;
  /// Parent Category.
  Category *mergeCatDestCat;

  // Autosave variables.
  guint timeoutID; ///< Handle for timeout function.
  guint idleID; ///< Handle for idle function.
  gboolean idleScheduled; ///< True if the idle function is scheduled.
};

#endif
