/*
 * $Id: QualResObject.h,v 1.7 2003/03/27 03:31:09 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifndef QUAL_RES_OBJECT_H
#define QUAL_RES_OBJECT_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <map>

#include "qualres.h"
#include "qualresUtility.h"
#include "SaveObject.h"

/** Base class for qualres domain classes.

    This class implements the basic characteristcs and error handling
    of all qualres objects involved in Qualitative Analysis (e.g.
    Document, Category).

    All access functions of this and derived classes return a pointer
    to const.  For example, getName returns (const string *), pointing
    to the name field.  The object is responsible for this memory; the
    calling program should never free the memory pointed to by this
    pointer.
*/
class QualResObject : public SaveObject
{
public:
  /** Types of QualRes objects.

      Each object stores a type value in its objType field when created.
      This value is used for run-time identification.
  */
  enum Type
  {
    PROJECT = 0,
    DOCUMENT,
    CATEGORY,
    SET,
    MEMO,
    XREF,
    NARRATIVE_DOCUMENT,
    FRAME,
    NTYPES
  };

  /** Relation of QualRes objects to create.

      When creating a Project Component, this enumeration specifies
      whether to create a sibling or child of the selected component.
  */
  enum Relationship
  {
    CHILD = 0,
    SIBLING,
    NRELATIONSHIPS
  };

  /** Tree view column numbers.
   */
  enum Column
  {
    SHORT_NAME = 0,
    DESCRIPTION,
    FULL_NAME,
    ADDRESS,
    FG_COLOR,
    BG_COLOR,
    FONT,
    N_DOCS,
    CHECK,
    N_COLUMNS
  };
    
  QualResObject();
  virtual ~QualResObject();

  void pendingSetName(const string &s);
  void commitName();
  virtual void uncommitName();

  /** Get the object's name.

      \return Object's name
  */
  const string &getName() const { return(name); }

  string getPrettyName() const;
  string getShortName();
  string getParentsName();

  void setDesc(const string &);

  /** Get the object's description property.

      \return Object's description
  */
  const string &getDesc() const { return(desc); }

  /** Get the object type.

      \return Object's type
  */
  enum Type getType() const { return(objType); }

  /// Inform the object that a component is about to be deleted.
  virtual void pendingDelete(QualResObject *) = 0;

  /// Inform the object that a component is about to be cut.
  virtual void pendingCut(QualResObject *) = 0;

  /// Inform the object that a component is about to be pasted.
  virtual void pendingPaste(QualResObject *) {}

  /// Inform the object that a component's properties have changed.
  virtual void propertiesHaveChanged(QualResObject *) = 0;

  // GUI methods.
  GtkWidget *createTreeView(GtkTreeStore *);
  GtkWidget *createCatTreeView(GtkTreeStore *);
  void addNDocsColumn(GtkWidget *, gint);
  void addCheckColumn(GtkWidget *, gint, string);

  /** Change the object's properties. */
  virtual bool changeProperties(GtkWindow *) = 0;
  virtual void createWindow(GtkWidget *) = 0; ///< Create the main window
  virtual void showWindow();
  /// Get main window widget.
  GtkWidget *getMainWindow() const { return(mainWindow); }

  /** Hide the window. */
  virtual void hideWindow()
    { if (mainWindow != NULL) gtk_widget_hide(mainWindow); }

private:
  string name; ///< Object name
  string desc; ///< Object description
  bool nameChanged; ///< True if there is a pending name change
  string oldName; ///< Original name
protected:
  Type objType; ///< Type of class
  string objName; ///< Name of class
  // Widgets.
  GtkWidget *mainWindow; ///< Widget main window
  GtkWidget *propWindow; ///< Properties window
  GtkWidget *compName; ///< Properties component name
  GtkTextBuffer *compDesc; ///< Properties component description
};

#endif
