/*
 * $Id: Set.cc,v 1.9 2003/04/16 14:11:47 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the qualres Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Set.h"

// Callbacks.
static void setInsertDocument_cb(GtkWidget*, gpointer);
static void setInsertCategory_cb(GtkWidget*, gpointer);
static void setInsertSet_cb(GtkWidget*, gpointer);
static void setRemoveDocument_cb(GtkWidget*, gpointer);
static void setRemoveCategory_cb(GtkWidget*, gpointer);
static void setRemoveSet_cb(GtkWidget*, gpointer);
static gboolean setButtonPress_cb(GtkWidget *, GdkEventButton *, gpointer);

/** Create empty Set.
 */
Set::Set()
{
  init();
}

/** Create Set initialized from XML element.
 */
Set::Set(GdomeElement *)
{
  init();
}

/** Initialize the object.
 */
void
Set::init()
{
  objType = QualResObject::SET;
  xmlTag = SET_TAG;
  memberDocStoreLoaded = FALSE;
  memberCatStoreLoaded = FALSE;
  memberSetStoreLoaded = FALSE;
  createWindow(NULL);
}

/** Destructor.
 */
Set::~Set()
{}

/** Show the main window.
 */
void
Set::showWindow()
{
  // Load the store.
  if (!memberDocStoreLoaded)
    loadSetMemberStore(QualResObject::DOCUMENT);
  if (!memberCatStoreLoaded)
    loadSetMemberStore(QualResObject::CATEGORY);
  if (!memberSetStoreLoaded)
    loadSetMemberSetStore();

  QualResObject::showWindow();
}

/** Inform the object that a component is about to be deleted.

    \param comp is the component
*/
void
Set::pendingDelete(QualResObject *comp)
{
  // If this Set is to be Deleted, hide the Selection Windows.
  if ((Set *) comp == this)
    {
      hideAllSelectWindows();
      return;
    }

  removeElement((Component *) comp);
}

/** Inform the object that a component is about to be cut.

    \param comp is the component
*/
void
Set::pendingCut(QualResObject *comp)
{
  pendingDelete(comp);
}

/** Inform the object that a component's properties have changed.

    \param comp is the component
*/
void
Set::propertiesHaveChanged(QualResObject *comp)
{
  Document *doc = (Document *) comp;
  Category *cat = (Category *) comp;

  GtkTreeIter otherSetIter;
  gint i, nMemberSets;

  // Is it a Document?
  if (memberDocMap.count(doc) != 0)
    loadSetMemberStore(QualResObject::DOCUMENT);
  // A Category?
  else if (memberCatMap.count(cat) != 0)
    loadSetMemberStore(QualResObject::CATEGORY);

  // Update the Set store.
  nMemberSets =
    gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
				   NULL);
  for (i = 0; i < nMemberSets; i++)
    {
      // Get the iterator for the Set row.
      gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
				    &otherSetIter,
				    NULL,
				    i);

      // Update the store for this Set row.
      updateComponentsInSetStoreSet(&otherSetIter);
    }
}

/** Get the text to print.

    \returns a pointer to allocated memory containing the text to print.
    This memory must be freed by g_free().

    This method returns a pointer to memory containing the list of
    Categories.

*/
gchar *
Set::getTextToPrint() const
{
  string buffer;

  // Address of Components.
  vector<Component *> comp;

  // Get the names of each Document
  comp.clear();
  gtk_tree_model_foreach(GTK_TREE_MODEL(memberDocStore),
			 loadComponentAddressesFromStore,
			 (gpointer) &comp);

  if (comp.size() != 0)
    buffer += "Member Documents--\n\n";

  // Load names and descriptions.
  addNamesAndDescriptions(buffer, comp);

  // Get the names of each Category.
  comp.clear();
  gtk_tree_model_foreach(GTK_TREE_MODEL(memberCatStore),
			 loadComponentAddressesFromStore,
			 (gpointer) &comp);

  if (comp.size() != 0)
    buffer += "Member Categories--\n\n";

  // Load names and descriptions.
  addNamesAndDescriptions(buffer, comp);

  return(g_strdup(buffer.c_str()));
}

/** Add the formatted names and descriptions of member Components.

    \param buffer is the variable to add the names and description to

    \param comp is the container of Component addresses

*/
void
Set::addNamesAndDescriptions(string &buffer, vector<Component *> &comp) const
{
  gint indx;
  string descBuffer;
  vector<Component *>::iterator iComp;

  for (iComp = comp.begin(); iComp != comp.end(); iComp++)
    {
      buffer += PRINT_COMPONENT_INDENT + (*iComp)->getPrettyName() + "\n";
      descBuffer = (*iComp)->getDesc();

      // Add padding before description.
      indx = 0;
      while ((indx = descBuffer.find(NEWLINE_CHAR, indx)) != string::npos)
	descBuffer.insert(++indx, PRINT_DESCRIPTION_PAD);

      if (descBuffer != "")
	buffer += PRINT_COMPONENT_INDENT PRINT_DESCRIPTION_PAD +
	  descBuffer + "\n\n";
      else
	buffer += "\n";
    }
}

/** Get the Component name labeled with "Category:  ".

    \returns left header of print output (free with g_free)
**/
gchar *
Set::getLabeledName() const
{
  return(g_strconcat("Set:  ", (getPrettyName()).c_str(), NULL));
}

/** Remove an element from all windows, if it is a member.

    \param comp is the element to remove.  The element is not destroyed.

    This method removes all occurances of a Set member.  This method
    should be called when the Set member is being Cut or Deleted.

*/
void
Set::removeElement(Component *comp)
{
  gboolean removedComponent = false;
  Document *doc = (Document *) comp;
  Category *cat = (Category *) comp;
  Set *set = (Set *) comp;
  map<const string, Set *>::iterator iSet;

  // Is it a Document?
  if (memberDocMap.count(doc) != 0)
    {
      memberDocMap.erase(doc);
      loadSetMemberStore(QualResObject::DOCUMENT);
      removedComponent = true;
    }
  // A Category?
  else if (memberCatMap.count(cat) != 0)
    {
      memberCatMap.erase(cat);
      loadSetMemberStore(QualResObject::CATEGORY);
      removedComponent = true;
    }
  // A Set?
  else if (memberSetMap.count(set) != 0)
    {
      memberSetMap.erase(set);
      removedComponent = true;
    }

  // Remove the Component from the Set store.
  deleteMemberFromSetStore(comp);

  // Inform all Sets that this Set removed a Component.
  if (removedComponent)
    {
      for (iSet = (*projectSetMap).begin();
	   iSet != (*projectSetMap).end();
	   iSet++)
	(iSet->second)->setChangedMember(this, comp);
    }
}

/** Load the names of all the members in a Set.

    \param memberNames is the container to which this method adds the
    names of the member Project Components

    \param type is the type of the component

*/
void
Set::loadMemberNames(map<const string, string> *memberNames,
		     QualResObject::Type type)
{
  map<Component *, Component *> *memberMap;
  map<Component *, Component *>::iterator iMember;
  map<Set *, Set *>::iterator iSet;
  string name;

  switch (type)
    {
    case QualResObject::DOCUMENT:
      memberMap = (map<Component *, Component *> *) &memberDocMap;
      break;
    case QualResObject::CATEGORY:
      memberMap = (map<Component *, Component *> *) &memberCatMap;
      break;
    default:
      g_assert_not_reached();
      return;
    }

  // Load all member Documents or Categories.
  for (iMember = memberMap->begin(); iMember != memberMap->end(); iMember++)
    {
      name = (iMember->second)->getName();
      (*memberNames)[name] = name;
    }

  // Load all Components that are members of member Sets.
  for (iSet = memberSetMap.begin(); iSet != memberSetMap.end(); iSet++)
    (iSet->second)->loadMemberNames(memberNames, type);
}

/** Check if an element is a member.

    \param comp is the element to check.

    \return true if elem is a member, false otherwise

*/
bool
Set::member(Component *comp)
{
  map<Document *, Document *>::iterator iDoc;
  map<Category *, Category *>::iterator iCat;
  map<Set *, Set *>::iterator iSet;

  // Check the member Documents.
  for (iDoc = memberDocMap.begin(); iDoc != memberDocMap.end(); iDoc++)
    {
      if (comp == iDoc->second)
	return(true);
    }

  // Check the member Categories.
  for (iCat = memberCatMap.begin(); iCat != memberCatMap.end(); iCat++)
    {
      if (comp == iCat->second)
	return(true);
    }

  // Check the member Sets.
  for (iSet = memberSetMap.begin(); iSet != memberSetMap.end(); iSet++)
    {
      // Check if this member Set is the element.
      if (comp == iSet->second)
	return(true);

      // Check if this member Set contains the element.
      if ((iSet->second)->member(comp))
	return(true);
    }

  // Not a member.
  return(false);
}

/** Write to an XML element.

    \param xmlDoc is the XML document

    \return a pointer to the XML element

    The objects creates an XML element, and writes its persistent data
    to it.
*/
GdomeElement *
Set::writeXML(GdomeDocument *xmlDoc)
  throw(FileException)
{
  GdomeElement *elem;
  GdomeText *descNode;
  GdomeException exc;
  GdomeDOMString *tempStr;

  // Create element.
  tempStr = gdome_str_mkref(SET_TAG);
  elem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
  gdome_str_unref(tempStr);
  if (elem == NULL)
    throw FileException();

  try
    {
      // Create and write the name element.
      setTextInElement(xmlDoc, elem,
		       NAME_TAG,
		       (gchar *) (getName()).c_str());

      // Create and write the description element.
      setTextInElement(xmlDoc, elem,
		       DESCRIPTION_TAG,
		       (gchar *) (getDesc()).c_str());

      // Save all Project Component names.
      writeMembersXML(xmlDoc, elem, QualResObject::DOCUMENT);
      writeMembersXML(xmlDoc, elem, QualResObject::CATEGORY);
      writeMembersXML(xmlDoc, elem, QualResObject::SET);
    }
  catch(FileException)
    {
      gdome_el_unref(elem, &exc);
      throw FileException();
    }

  setSaved();
  return(elem);
}

/** Write all members of a Component type to an XML element.

    \param xmlDoc is the XML document

    \param setElem is the Set element

    \param type is the type of the component

*/
void
Set::writeMembersXML(GdomeDocument *xmlDoc,
	      GdomeElement *setElem,
	      QualResObject::Type type)
  throw(FileException)
{
  GdomeElement *memberElem;
  GdomeNode *result;
  GdomeDOMString *tempStr;
  GdomeException exc;
  map<Component *, Component *> *memberMap;
  map<Component *, Component *>::iterator iMember;
  gchar *tag;

  switch (type)
    {
    case QualResObject::DOCUMENT:
      memberMap = (map<Component *, Component *> *) &memberDocMap;
      tag = MEMBER_DOCUMENT_TAG;
      break;
    case QualResObject::CATEGORY:
      memberMap = (map<Component *, Component *> *) &memberCatMap;
      tag = MEMBER_CATEGORY_TAG;
      break;
    case QualResObject::SET:
      memberMap = (map<Component *, Component *> *) &memberSetMap;
      tag = MEMBER_SET_TAG;
      break;
    }

  for (iMember = memberMap->begin();
       iMember != memberMap->end();
       iMember++)
    {
      // Create the XML element for the member Document.
      tempStr = gdome_str_mkref(tag);
      memberElem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
      gdome_str_unref(tempStr);
      if (memberElem == NULL)
	throw FileException();

      try
	{
	  // Set the name.
	  setTextInElement(xmlDoc,
			   memberElem,
			   NAME_TAG,
			   (gchar *) ((iMember->second)->getName()).c_str());
	}
      catch(FileException)
	{
	  gdome_el_unref(memberElem, &exc);
	  throw FileException();
	}
	  

      // Append memberElem to the childs list of the Set element.
      result = gdome_el_appendChild(setElem,
				    (GdomeNode *) memberElem,
				    &exc);
      if (result != (GdomeNode *) memberElem)
	{
	  gdome_el_unref(memberElem, &exc);
	  throw FileException();
	}
      gdome_el_unref(memberElem, &exc);
    }
}

/** Read from an XML element.

    \param elem is the XML element containing the object data

    The member Set map is NOT initialized, because all Sets must be
    initialized.  The member Set names are stored, and the member Set
    map is initialized when loadSetMembers is invoked.

*/
bool
Set::readXML(GdomeElement *elem)
  throw(FileException)
{
  GdomeNodeList *memberElems;
  GdomeElement *memberElem;
  GdomeException exc;
  GdomeDOMString *tempStr;
  string compType, memberName;
  gint nComponents, i;

  // Load the name.
  pendingSetName(getTextFromElement(elem, NAME_TAG));
  commitName();

  // Load the description.
  setDesc(getTextFromElement(elem, DESCRIPTION_TAG));

  // Get the reference to all the member elements.
  memberElems = gdome_el_childNodes(elem, &exc);
  if (memberElems == NULL)
    throw FileException();
  nComponents = gdome_nl_length(memberElems, &exc);

  // Clear the names of the member Sets.
  memberSetNames.clear();

  // Load all member names.
  for (i = 0; i < nComponents; i++)
    {
      // Obtain the next member name element
      memberElem = (GdomeElement *)gdome_nl_item(memberElems, i, &exc);
      if (memberElem == NULL)
	{
	  gdome_nl_unref(memberElems, &exc);
	  throw FileException();
	}

      // Check that it is an element.
      if (gdome_el_nodeType(memberElem, &exc) == GDOME_ELEMENT_NODE)
	{
	  // It is.  Get the tag name.
	  tempStr = gdome_el_tagName(memberElem, &exc);
	  if (exc)
	    {
	      gdome_el_unref (memberElem, &exc);
	      gdome_nl_unref(memberElems, &exc);
	      throw FileException();
	    }
	  compType = tempStr->str;
	  gdome_str_unref(tempStr);

	  // Load the appropriate member map.
	  try
	    {
	      memberName = getTextFromElement(memberElem, NAME_TAG);
	    }
	  catch(FileException)
	    {
	      gdome_el_unref (memberElem, &exc);
	      gdome_nl_unref(memberElems, &exc);
	      throw FileException();
	    }

	  if (compType == MEMBER_DOCUMENT_TAG)
	    memberDocMap[(*projectDocMap)[memberName]] =
	      (*projectDocMap)[memberName];
	  else if (compType == MEMBER_CATEGORY_TAG)
	    memberCatMap[(*projectCatMap)[memberName]] =
	      (*projectCatMap)[memberName];
	  else if (compType == MEMBER_SET_TAG)
	    // The member Set might not have been read in.  Save the
	    // name, and assign the pointer later.
	    memberSetNames.push_back(memberName);
	}

      gdome_el_unref (memberElem, &exc);
    }

  gdome_nl_unref(memberElems, &exc);

  setSaved();
  return(TRUE);
}

/** Load the Set member map from the names.

    This method must be called after initializing a Set from an XML
    element.  The map of member Sets is initialized from the names
    loaded by readXML.
*/
void
Set::loadSetMembers()
{
  Set *set;
  vector<string>::iterator iName;

  for (iName = memberSetNames.begin();
       iName != memberSetNames.end();
       iName++)
    {
      set = (*projectSetMap)[*iName];
      memberSetMap[set] = set;
    }

  memberSetNames.clear();
}

/** Inform this Set that a different Set added or removed a member.

    \param otherSet is the Set that added a member Component.

    \param comp is the Component that was added or removed.

*/
void
Set::setChangedMember(Set *otherSet, Component *comp)
{
  GtkTreeIter otherSetIter;
  gint i, nMemberSets;

  // This set has already updated the store.
  if (otherSet == this)
    return;

  // Update the otherSet row.
  nMemberSets =
    gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
				   NULL);
  for (i = 0; i < nMemberSets; i++)
    {
      // Get the iterator for the Set row.
      gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
				    &otherSetIter,
				    NULL,
				    i);

      // Update the store for this Set row.
      updateComponentsInSetStoreSet(&otherSetIter);
    }
}

/** Change the object's properties.

    \param parent is the parent window

 */
bool
Set::changeProperties(GtkWindow *parent)
{
  GtkWidget *table;
  GtkAttachOptions labelAttach = GTK_SHRINK, textAttach = GTK_FILL;
  GtkWidget *view, *scrolledView;
  GtkTextIter start, end;
  bool status;
  gint result;
  string sProp;

  // Create the properties window.
  if (propWindow == NULL)
    {
      propWindow = gtk_dialog_new_with_buttons("Properties",
					       parent,
					       GTK_DIALOG_MODAL,
					       GTK_STOCK_OK,
					       GTK_RESPONSE_ACCEPT,
					       GTK_STOCK_CANCEL,
					       GTK_RESPONSE_REJECT,
					       NULL);
      table = gtk_table_new(2, 2, TRUE);


      // Add label and text entry for the name.
      gtk_table_attach(GTK_TABLE(table),
		       gtk_label_new(_("Set name:  ")),
		       0, 1, 0, 1,
		       labelAttach, labelAttach,
		       5, 5);
      compName = gtk_entry_new();
      gtk_table_attach(GTK_TABLE(table),
		       compName,
		       1, 2, 0, 1,
		       textAttach, textAttach,
		       5, 5);

      // Add label and text buffer for the description.
      gtk_table_attach(GTK_TABLE(table),
		       gtk_label_new(_("Set description:  ")),
		       0, 1, 1, 2,
		       labelAttach, labelAttach,
		       5, 5);

      // Create view, and put it into a scrolled window.
      view = gtk_text_view_new();
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), view);

      compDesc = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
      gtk_table_attach(GTK_TABLE(table),
		       scrolledView,
		       1, 2, 1, 2,
		       textAttach, textAttach,
		       5, 5);

      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(propWindow)->vbox),
			table);
    }

  // Load the properties.
  sProp = getShortName();
  gtk_entry_set_text(GTK_ENTRY(compName), sProp.c_str());
  gtk_text_buffer_get_bounds(compDesc, &start, &end);
  gtk_text_buffer_delete(compDesc, &start, &end);
  sProp = getDesc();
  gtk_text_buffer_set_text(compDesc, sProp.c_str(), -1);

  // Show the properties window.
  gtk_widget_show_all(propWindow);
  result = gtk_dialog_run(GTK_DIALOG(propWindow));
  gtk_widget_hide(propWindow);

  // Load new properties.
  if (result == GTK_RESPONSE_ACCEPT)
    {
      sProp = gtk_entry_get_text(GTK_ENTRY(compName));
      pendingSetName(getParentsName() + sProp);
      gtk_text_buffer_get_bounds(compDesc, &start, &end);
      setDesc(gtk_text_buffer_get_text(compDesc, &start, &end, FALSE));
      status = true;
    }
  else
    status = false;

  return(status);
}

/** Create the window to update and view a Set.

    \param parent is the parent window

 */
void
Set::createWindow(GtkWidget *parent)
{
  /* Menus */
  static GnomeUIInfo fileMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Print members"),
     N_("Print the member Documents and Categories"),
     (gpointer) print_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Close window"),
     N_("Close the Set window"),
     (gpointer) close_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CLOSE,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo editMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Insert Document"),
     N_("Select Documents to insert from a popup list"),
     (gpointer) setInsertDocument_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Insert Category"),
     N_("Select Categories to insert from a popup list"),
     (gpointer) setInsertCategory_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Insert Set"),
     N_("Select Sets to insert from a popup list"),
     (gpointer) setInsertSet_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Remove Document"),
     N_("Remove the selected Document"),
     (gpointer) setRemoveDocument_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_DELETE,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Remove Category"),
     N_("Remove the selected Category"),
     (gpointer) setRemoveCategory_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_DELETE,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Remove Set"),
     N_("Remove the selected Set"),
     (gpointer) setRemoveSet_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_DELETE,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo help_menu [] = {
    GNOMEUIINFO_HELP ("Set"),
  
    GNOMEUIINFO_END
  };

  static GnomeUIInfo menu [] = {
    { GNOME_APP_UI_SUBTREE_STOCK, N_("File"), NULL, fileMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    { GNOME_APP_UI_SUBTREE_STOCK, N_("Edit"), NULL, editMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    GNOMEUIINFO_MENU_HELP_TREE(help_menu),
    GNOMEUIINFO_END
  };

  GtkWidget *status, *vbox, *frame;
  GtkWidget *button, *scrolledView;
  GtkWidget *hPaneLeft, *hPaneRight;
  
  string sProp;

  // Create the main window.
  if (mainWindow == NULL)
    {
      // Create widgets.
      mainWindow = gnome_app_new("Set", _("Set a"));
      gtk_window_set_default_size(GTK_WINDOW(mainWindow),
				  SET_XSIZE,
				  SET_YSIZE);

      status = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_NEVER);
      frame = gtk_frame_new(NULL);
      gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);

      gtk_signal_connect(GTK_OBJECT(mainWindow), "delete-event",
			 GTK_SIGNAL_FUNC(hide_cb), this);

      // Create the models and view for the member trees.
      memberDocStore = gtk_tree_store_new(4,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_POINTER);
      memberDocView = createTreeView(memberDocStore);
      g_signal_connect(G_OBJECT(memberDocView),
		       "button-press-event",
		       G_CALLBACK(setButtonPress_cb),
		       this);

      memberCatStore = gtk_tree_store_new(QualResObject::N_COLUMNS,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_POINTER,
					  GDK_TYPE_COLOR,
					  GDK_TYPE_COLOR,
					  G_TYPE_STRING,
					  G_TYPE_UINT,
					  G_TYPE_BOOLEAN);
      memberCatView = createCatTreeView(memberCatStore);
      addNDocsColumn(memberCatView, NDOCS_COLUMN_NUMBER);
      g_signal_connect(G_OBJECT(memberCatView),
		       "button-press-event",
		       G_CALLBACK(setButtonPress_cb),
		       this);

      memberSetStore = gtk_tree_store_new(QualResObject::N_COLUMNS,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_POINTER,
					  GDK_TYPE_COLOR,
					  GDK_TYPE_COLOR,
					  G_TYPE_STRING,
					  G_TYPE_UINT,
					  G_TYPE_BOOLEAN);

      memberSetView = createCatTreeView(memberSetStore);
      g_signal_connect(G_OBJECT(memberSetView),
		       "button-press-event",
		       G_CALLBACK(setButtonPress_cb),
		       this);

      // Pack the views.
      vbox = gtk_vbox_new(FALSE, 5);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_label_new(_("Member Documents")),
			 FALSE, FALSE, 0);
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), memberDocView);
      gtk_box_pack_start(GTK_BOX(vbox),
			 scrolledView,
			 TRUE, TRUE, 0);

      hPaneLeft = gtk_hpaned_new();
      gtk_paned_pack1(GTK_PANED(hPaneLeft), vbox, FALSE, TRUE);

      vbox = gtk_vbox_new(FALSE, 5);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_label_new(_("Member Categories")),
			 FALSE, FALSE, 0);
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), memberCatView);
      gtk_box_pack_start(GTK_BOX(vbox),
			 scrolledView,
			 TRUE, TRUE, 0);

      gtk_paned_pack2(GTK_PANED(hPaneLeft), vbox, TRUE, TRUE);
      gtk_paned_set_position(GTK_PANED(hPaneLeft), SET_XSIZE / 3);

      vbox = gtk_vbox_new(FALSE, 5);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_label_new(_("Member Sets")),
			 FALSE, FALSE, 0);
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), memberSetView);
      gtk_box_pack_start(GTK_BOX(vbox),
			 scrolledView,
			 TRUE, TRUE, 0);

      hPaneRight = gtk_hpaned_new();
      gtk_paned_pack1(GTK_PANED(hPaneRight), hPaneLeft, FALSE, TRUE);
      gtk_paned_pack2(GTK_PANED(hPaneRight), vbox, TRUE, TRUE);
      gtk_paned_set_position(GTK_PANED(hPaneRight), 2 * SET_XSIZE / 3);

      // Add to application window.
      gnome_app_set_contents(GNOME_APP(mainWindow), frame);
      gnome_app_set_statusbar(GNOME_APP(mainWindow), status);
      gnome_app_create_menus_with_data(GNOME_APP(mainWindow), menu, this);
      gnome_app_install_menu_hints(GNOME_APP(mainWindow), menu);

      // Add the widgets.
      gtk_container_add(GTK_CONTAINER(frame), hPaneRight);
    }
}

/** Load the store of the Set members.

    \param type specifies which model to update (DOCUMENT or CATEGORY)

    The model is first cleared, and loaded with the names of the
    Set contents
*/
void
Set::loadSetMemberStore(QualResObject::Type type)
{
  // Pointer to Component in question.
  Component *comp;

  // Pointer to Project component container.
  map<const string, Component *> *projectMap;

  // Pointer to members.
  map<Component *, Component *> *memberMap;
  map<Component *, Component *>::iterator iMember;

  // Tree model.
  GtkTreeStore *store;
  GtkTreeIter iter, parentIter;

  // Index.
  gint i;

  // Point to the store.
  switch (type)
    {
    case QualResObject::DOCUMENT:
      memberDocStoreLoaded = TRUE;
      memberMap = (map<Component *, Component *> *) &memberDocMap;
      projectMap = (map<const string, Component *> *) projectDocMap;
      store = memberDocStore;
      break;
    case QualResObject::CATEGORY:
      memberCatStoreLoaded = TRUE;
      memberMap = (map<Component *, Component *> *) &memberCatMap;
      projectMap = (map<const string, Component *> *) projectCatMap;
      store = memberCatStore;
      break;
    default:
      g_assert_not_reached();
      return;
    }

  // Clear the store.
  gtk_tree_store_clear(store);
      
  // Add rows.
  for (iMember = memberMap->begin(); iMember != memberMap->end(); iMember++)
    {
      comp = iMember->second;

      // Store it.
      gtk_tree_store_append(store, &iter, NULL);

      gtk_tree_store_set(store, &iter,
			 QualResObject::SHORT_NAME,
			 (comp->getPrettyName()).c_str(),
			 QualResObject::DESCRIPTION, (comp->getDesc()).c_str(),
			 QualResObject::FULL_NAME, (comp->getName()).c_str(),
			 QualResObject::ADDRESS, comp,
			 -1);

      // For Categories, update the text markup and number of Documents.
      if (type == QualResObject::CATEGORY)
	{
	  Category *cat = (Category *) comp;
	  gtk_tree_store_set(store, &iter,
			     QualResObject::FG_COLOR, cat->getFgColor(),
			     QualResObject::BG_COLOR, cat->getBgColor(),
			     QualResObject::FONT, (*(cat->getFont())).c_str(),
			     QualResObject::N_DOCS, cat->nDocuments(),
			     -1);
	}
    }
}

/** Load the Set store of the Set members.

    The model is first cleared, and loaded with the names of the
    Set contents
*/
void
Set::loadSetMemberSetStore()
{
  // Pointer to member Sets.
  map<Set *, Set *>::iterator iSet;

  // Clear the store.
  gtk_tree_store_clear(memberSetStore);
      
  // Add rows.
  for (iSet = memberSetMap.begin(); iSet != memberSetMap.end(); iSet++)
      loadSetRow(iSet->second);
}

/** Load one Set into the member Set store.

    \param set is the Set to load

*/
void
Set::loadSetRow(Set *set)
{
  // Set members.
  map<const string, string> docNames, catNames;
  map<const string, string>::iterator iName;
  gint nDocs, nCats;
  Document *doc;
  Category *cat;

  GtkTreeIter setIter, labelIter, compIter;

  // Store it.
  gtk_tree_store_append(memberSetStore, &setIter, NULL);

  // Load the member names.
  gtk_tree_store_set(memberSetStore, &setIter,
		     QualResObject::SHORT_NAME, (set->getPrettyName()).c_str(),
		     QualResObject::DESCRIPTION, (set->getDesc()).c_str(),
		     QualResObject::FULL_NAME, (set->getName()).c_str(),
		     QualResObject::ADDRESS, set,
		     -1);

  // Load the Document names.
  set->loadMemberNames(&docNames, QualResObject::DOCUMENT);
  nDocs = docNames.size();
  if (nDocs > 0)
    {
      // Load the label row.
      gtk_tree_store_append(memberSetStore, &labelIter, &setIter);
      gtk_tree_store_set(memberSetStore, &labelIter,
			 QualResObject::SHORT_NAME, _(SET_DOCUMENT_LABEL),
			 QualResObject::ADDRESS, NULL,
			 -1);

      // Load the member rows.
      for (iName = docNames.begin(); iName != docNames.end(); iName++)
	{
	  doc = (*projectDocMap)[iName->second];
	  gtk_tree_store_append(memberSetStore, &compIter, &labelIter);
	  gtk_tree_store_set(memberSetStore, &compIter,
			     QualResObject::SHORT_NAME,
			     (doc->getPrettyName()).c_str(),
			     QualResObject::DESCRIPTION,
			     (doc->getDesc()).c_str(),
			     QualResObject::FULL_NAME,
			     (doc->getName()).c_str(),
			     QualResObject::ADDRESS, doc,
			     -1);
	}
    }

  // Load the member names.
  set->loadMemberNames(&catNames, QualResObject::CATEGORY);
  nCats = catNames.size();
  if (nCats > 0)
    {
      // Load the label row.
      gtk_tree_store_append(memberSetStore, &labelIter, &setIter);
      gtk_tree_store_set(memberSetStore, &labelIter,
			 QualResObject::SHORT_NAME, _(SET_CATEGORY_LABEL),
			 QualResObject::ADDRESS, NULL,
			 -1);

      // Load the member rows.
      for (iName = catNames.begin(); iName != catNames.end(); iName++)
	{
	  cat = (*projectCatMap)[iName->second];
	  gtk_tree_store_append(memberSetStore, &compIter, &labelIter);
	  gtk_tree_store_set(memberSetStore, &compIter,
			     QualResObject::SHORT_NAME,
			     (cat->getPrettyName()).c_str(),
			     QualResObject::DESCRIPTION,
			     (cat->getDesc()).c_str(),
			     QualResObject::FULL_NAME,
			     (cat->getName()).c_str(),
			     QualResObject::ADDRESS, cat,
			     QualResObject::FG_COLOR, cat->getFgColor(),
			     QualResObject::BG_COLOR, cat->getBgColor(),
			     QualResObject::FONT, (*(cat->getFont())).c_str(),
			     -1);
	}
    }
}

/** Update the Components for a Set in the memberSetStore.

    \param setIter points to the Set in memberSetStore.

*/
void
Set::updateComponentsInSetStoreSet(GtkTreeIter *setIter)
{
  Set *set;

  // Flags.
  gboolean haveCatLabel, haveDocLabel, docExpanded;

  // Temporary.
  gchar *label;
  string sLabel;

  // Set members.
  map<const string, string> names;
  gint i, nLabels, nDocs, nCats, nRows;
  Document *doc;
  Category *cat;

  GtkTreeIter labelIter, docLabelIter, catLabelIter, compIter;
  GtkTreePath *path;

  // Get the Set address.
  gtk_tree_model_get(GTK_TREE_MODEL(memberSetStore),
		     setIter,
		     Set::ADDRESS, &set,
		     -1);

  // Get the label iterators.
  nLabels = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
					   setIter);
  haveCatLabel = false;
  haveDocLabel = false;
  for (i = 0; i < nLabels; i++)
    {
      // Get the first label.
      gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
				    &labelIter,
				    setIter,
				    i);

      gtk_tree_model_get(GTK_TREE_MODEL(memberSetStore),
			 &labelIter,
			 Set::SHORT_NAME, &label,
			 -1);

      sLabel = label;
      g_free(label);
      if (sLabel == SET_CATEGORY_LABEL)
	{
	  haveCatLabel = true;
	  catLabelIter = labelIter;
	}
      else if (sLabel == SET_DOCUMENT_LABEL)
	{
	  haveDocLabel = true;
	  docLabelIter = labelIter;
	}
      else
	{
	  g_assert_not_reached();
	  return;
	}
    }

  // Load the Document names.
  set->loadMemberNames(&names, QualResObject::DOCUMENT);

  // Replace the Document members.
  updateLabelComponents(&names,
			&docLabelIter,
			haveDocLabel,
			QualResObject::DOCUMENT,
			setIter);
  names.clear();

  // Load the Category names.
  set->loadMemberNames(&names, QualResObject::CATEGORY);

  // Replace the Category members.
  updateLabelComponents(&names,
			&catLabelIter,
			haveCatLabel,
			QualResObject::CATEGORY,
			setIter);
}

/** Replace the members of a Set.

    \param names is the names of the members of this set

    \param labelIter points to the label row of the store, if there is one

    \param iterValid is true if there is a label row

    \param type is the type of Component

    \param setIter points to the Set row of the store
*/
void
Set::updateLabelComponents(map<const string, string> *names,
			   GtkTreeIter *labelIter,
			   gboolean iterValid,
			   enum QualResObject::Type type,
			   GtkTreeIter *setIter)
{
  GtkTreePath *path;
  gboolean expanded; // True if the row is expanded
  gint nRows, nComps, i;
  GtkTreeIter compIter;
  gchar *labelName;
  Component *comp;
  map<const string, Component *> *projectMap;
  map<const string, string>::iterator iName;

  switch (type)
    {
    case QualResObject::DOCUMENT:
      projectMap = (map<const string, Component *> *) projectDocMap;
      labelName = _(SET_DOCUMENT_LABEL);
      break;
    case QualResObject::CATEGORY:
      projectMap = (map<const string, Component *> *) projectCatMap;
      labelName = _(SET_CATEGORY_LABEL);
      break;
    default:
      g_assert_not_reached();
      return;
    }

  // Find out whether the label rows are expanded.
  if (iterValid)
    {
      // Get the path for the label row.
      path = gtk_tree_model_get_path(GTK_TREE_MODEL(memberSetStore),
				     labelIter);
      // Get the expansion state.
      expanded = gtk_tree_view_row_expanded(GTK_TREE_VIEW(memberSetView),
					       path);
      // Free the path.
      gtk_tree_path_free(path);
    }
  else
    expanded = false;

  // Remove the original rows.
  if (iterValid)
    {
      nRows = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
					     labelIter);
      for (i = nRows - 1; i >= 0; i--)
	{
	  gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
					&compIter,
					labelIter,
					i);
	  gtk_tree_store_remove(memberSetStore, &compIter);
	}
    }

  // Load the Component names.
  nComps = names->size();
  if (nComps > 0)
    {
      // Load the label row.
      if (!iterValid)
	{
	  gtk_tree_store_append(memberSetStore, labelIter, setIter);
	  gtk_tree_store_set(memberSetStore, labelIter,
			     QualResObject::SHORT_NAME, labelName,
			     QualResObject::ADDRESS, NULL,
			     -1);
	}

      // Load the member rows.
      for (iName = names->begin(); iName != names->end(); iName++)
	{
	  comp = (*projectMap)[iName->second];
	  gtk_tree_store_append(memberSetStore, &compIter, labelIter);
	  gtk_tree_store_set(memberSetStore, &compIter,
			     QualResObject::SHORT_NAME,
			     (comp->getPrettyName()).c_str(),
			     QualResObject::DESCRIPTION,
			     (comp->getDesc()).c_str(),
			     QualResObject::FULL_NAME,
			     (comp->getName()).c_str(),
			     QualResObject::ADDRESS, comp,
			     -1);
	  // Update the text markup if the component is a Category.
	  if (type == QualResObject::CATEGORY)
	    {
	      Category *cat = (Category *) comp;
	      gtk_tree_store_set(memberSetStore, &compIter,
				 QualResObject::FG_COLOR, cat->getFgColor(),
				 QualResObject::BG_COLOR, cat->getBgColor(),
				 QualResObject::FONT,
				 (*(cat->getFont())).c_str(),
				 -1);
	    }
	}

      // Restore the state of the label.
      if (expanded)
	{
	  // Get the path for the label row.
	  path = gtk_tree_model_get_path(GTK_TREE_MODEL(memberSetStore),
					 labelIter);
	  // Expand it.
	  gtk_tree_view_expand_row(GTK_TREE_VIEW(memberSetView),
				   path,
				   TRUE);
	  // Free the path.
	  gtk_tree_path_free(path);
	}

    }
  else if (iterValid)
    // Remove label as there are no more Components.
    gtk_tree_store_remove(memberSetStore, labelIter);
}

/** Delete a member from the store of the Set members.

    \param comp is the address of the component to delete

*/
void
Set::deleteMemberFromSetStore(Component *comp)
{
  // Tree model iterators.
  GtkTreeIter setIter;
  gboolean validSetIter;

  // Component the row refers to.
  Component *rowComp;

  // Point to the first row.
  validSetIter = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(memberSetStore),
					  &setIter);
  // Check each Set row.
  while (validSetIter)
    {
      // Get the address.
      gtk_tree_model_get(GTK_TREE_MODEL(memberSetStore),
			 &setIter,
			 Set::ADDRESS, &rowComp,
			 -1);

      // Delete this?
      if (rowComp == comp)
	{
	  deleteTreeRow(memberSetStore, &setIter);
	  // We're done.
	  break;
	}

      // Delete the Component from the children.
      deleteMemberFromSetStoreSet(&setIter, comp);

      // Get the next Set row.
      validSetIter = gtk_tree_model_iter_next(GTK_TREE_MODEL(memberSetStore),
					 &setIter);
    }
}

/** Delete a Component from a Set in the memberSetStore.

    \param setIter points to the Set in memberSetStore.

    \param comp is the Component to delete
*/
void
Set::deleteMemberFromSetStoreSet(GtkTreeIter *setIter, Component *comp)
{
  GtkTreeIter labelIter, compIter;

  // Component the row refers to.
  Component *rowComp;

  // Number of children of each row.
  gint nSetChildren, nComp;

  // Index.
  gint i, j;

  nSetChildren = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
						setIter);
  for (i = nSetChildren - 1; i >= 0; i--)
    {
      // Get the iterator for the Document or Category label row.
      gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
				    &labelIter,
				    setIter,
				    i);
      // Get the number of member Documents or Categories.
      nComp =
	gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
				       &labelIter);
      for (j = nComp - 1; j >= 0; j--)
	{
	  // Get the iterator for the Document or Category row.
	  gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(memberSetStore),
					&compIter,
					&labelIter,
					j);

	  // Get the address.
	  gtk_tree_model_get(GTK_TREE_MODEL(memberSetStore),
			     &compIter,
			     Set::ADDRESS, &rowComp,
			     -1);

	  // Delete this?
	  if (rowComp == comp)
	    gtk_tree_store_remove(memberSetStore, &compIter);
	}

      // Remove the label if there are no members remaining.
      nComp = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(memberSetStore),
					     &labelIter);
      if (nComp == 0)
	gtk_tree_store_remove(memberSetStore, &labelIter);
    }
}

/** Insert a component into the Set

    \param type is the type of Component

*/
void
Set::insertComponent(QualResObject::Type type)
{
  // Variables used in creating the window.
  GtkWidget *view, *vbox, *scrolledView;
  GtkAdjustment *adjustment;
  GtkTreeSelection *selection;
  gint result;

  // List of selected Project Components.
  map<const string, string> selectedNames;
  map<const string, string>::iterator iName;

  GtkWidget **selectWindow, **selectView;
  GtkTreeStore *selectStore;

  map<Component *, Component *> *memberMap;
  map<const string, Component *> *projectMap;
  map<const string, Set *>::iterator iSet;
  Set *setToAdd;

  string prompt = "Choose a ", titlePrompt = "Select a ";
  string compType;

  // Error message.
  string dialogMessage;

  switch (type)
    {
    case QualResObject::DOCUMENT:
      compType = "Document";
      memberMap = (map<Component *, Component *> *) &memberDocMap;
      projectMap = (map<const string, Component *> *) projectDocMap;
      selectWindow = &selectDocWindow;
      selectView = &selectDocView;
      selectStore = selectDocStore;
      break;
    case QualResObject::CATEGORY:
      compType = "Category";
      memberMap = (map<Component *, Component *> *) &memberCatMap;
      projectMap = (map<const string, Component *> *) projectCatMap;
      selectWindow = &selectCatWindow;
      selectView = &selectCatView;
      selectStore = selectCatStore;
      break;
    case QualResObject::SET:
      compType = "Set";
      memberMap = (map<Component *, Component *> *) &memberSetMap;
      projectMap = (map<const string, Component *> *) projectSetMap;
      selectWindow = &selectSetWindow;
      selectView = &selectSetView;
      selectStore = selectSetStore;
      break;
    }
  titlePrompt += compType;
  prompt += compType;

  // Create the Component selection window.
  if (*selectWindow == NULL)
    {
      *selectWindow = gtk_dialog_new_with_buttons(titlePrompt.c_str(),
						  GTK_WINDOW(mainWindow),
						  GTK_DIALOG_MODAL,
						  GTK_STOCK_OK,
						  GTK_RESPONSE_ACCEPT,
						  GTK_STOCK_CANCEL,
						  GTK_RESPONSE_REJECT,
						  NULL);
      gtk_window_set_default_size(GTK_WINDOW(*selectWindow),
				  SELECT_COMPONENT_XSIZE,
				  SELECT_COMPONENT_YSIZE);

      // Create the widgets.
      if (type == QualResObject::CATEGORY)
	{
	  *selectView = createCatTreeView(selectStore);
	  addNDocsColumn(*selectView, NDOCS_COLUMN_NUMBER);
	}
      else
	*selectView = createTreeView(selectStore);

      // Allow multiple selections.
      selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(*selectView));
      gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

      // Pack the vbox.
      vbox = gtk_vbox_new(FALSE, 5);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_label_new(_(prompt.c_str())),
			 FALSE, FALSE, 0);
      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), *selectView);
      gtk_box_pack_start(GTK_BOX(vbox),
			 scrolledView,
			 TRUE, TRUE, 0);
				       
      // Add the view to the selection widget.
      gtk_container_add(GTK_CONTAINER(GTK_DIALOG(*selectWindow)->vbox),
			vbox);

      // Scroll selectCatView all the way to the left.
      gtk_widget_show_all(*selectWindow);
      adjustment =
	gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(scrolledView));
      gtk_adjustment_set_value(adjustment, adjustment->lower);
      gtk_adjustment_value_changed(adjustment);
    }

  // Show the selection window.
  gtk_widget_show_all(*selectWindow);
  result = gtk_dialog_run(GTK_DIALOG(*selectWindow));
  gtk_widget_hide(*selectWindow);

  // Load the new Component into the Set.
  if (result == GTK_RESPONSE_ACCEPT)
    {
      // Get the selected Component.
      selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(*selectView));

      // Get the names.
      selectedNames.clear();
      gtk_tree_selection_selected_foreach(selection,
					  addNamesFromView,
					  (gpointer) &selectedNames);

      // Add each Component.
      for (iName = selectedNames.begin();
	   iName != selectedNames.end();
	   iName++)
	{
	  // Go to the next name if this Component is already a member.
	  if ((*memberMap).count((*projectMap)[iName->second]) != 0)
	    continue;

	  // If we're inserting a Set, check that this Set is not a member
	  // of the Set to insert.
	  if (type == QualResObject::SET)
	    {
	      setToAdd = (*projectSetMap)[iName->second];
	      if (setToAdd->member(this) || (setToAdd == this))
		{
		  dialogMessage = "Cannot add " + setToAdd->getPrettyName() +
		    "; doing so would make " + this->getPrettyName() +
		    " contain itself, which does not make sense";
		  showMessage(dialogMessage, GTK_MESSAGE_ERROR, mainWindow);
		  continue;
		}
	    }

	  // Load the Component into the member map.
	  (*memberMap)[(*projectMap)[iName->second]] =
	    (*projectMap)[iName->second];

	  // Add the row to the store.
	  if (type == QualResObject::SET)
	    loadSetRow(setToAdd);
	  else
	    loadSetMemberStore(type);

	  // Inform all Sets that a member was added to this Set.
	  for (iSet = (*projectSetMap).begin();
	       iSet != (*projectSetMap).end();
	       iSet++)
	    (iSet->second)->setChangedMember(this,
					     (*projectMap)[iName->second]);
	  setChanged();
	}
    }
}

/** Remove the selected Component from the Set

    \param type is the type of Component

    This method should be called when a Set member is removed from the
    Set.

*/
void
Set::removeComponent(QualResObject::Type type)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkTreeStore *store;
  GtkWidget *view;
  GtkTreeSelection *selection;
  gchar *name;
  string compType;

  Component *comp;
  map<Component *, Component *> *memberMap;
  map<const string, Component *> *projectMap;
  map<const string, Set *>::iterator iSet;

  switch (type)
    {
    case QualResObject::DOCUMENT:
      compType = "Document";
      memberMap = (map<Component *, Component *> *) &memberDocMap;
      projectMap = (map<const string, Component *> *) projectDocMap;
      view = memberDocView;
      store = memberDocStore;
      break;
    case QualResObject::CATEGORY:
      compType = "Category";
      memberMap = (map<Component *, Component *> *) &memberCatMap;
      projectMap = (map<const string, Component *> *) projectCatMap;
      view = memberCatView;
      store = memberCatStore;
      break;
    case QualResObject::SET:
      compType = "Set";
      memberMap = (map<Component *, Component *> *) &memberSetMap;
      projectMap = (map<const string, Component *> *) projectSetMap;
      view = memberSetView;
      store = memberSetStore;
      break;
    }

  // Get the selected Project Component.
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));

  // Popup error box if there's no selection.
  if (!gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      showMessage("You must first select a " + compType + " to remove",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  gtk_tree_model_get(GTK_TREE_MODEL(store),
		     &iter,
		     Set::FULL_NAME, &name,
		     Set::ADDRESS, &comp,
		     -1);

  // Check that the component of the proper type was selected (for
  // the Set store, the user can select Documents or Categories).
  if (comp == NULL)
    {
      showMessage("You must first select a " + compType + " to remove",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  if (comp->getType() != type)
    {
      showMessage("You must first select a " + compType + " to remove",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  // Remove the Project Component from the Set.
  removeElement(comp);
  memberMap->erase((*projectMap)[name]);

  // Delete the row in the Component store.
  if (type == QualResObject::SET)
    deleteMemberFromSetStore((*projectMap)[name]);
  else
    loadSetMemberStore(type);

  // Inform all Sets that this Set removed a Component.
  for (iSet = (*projectSetMap).begin();
       iSet != (*projectSetMap).end();
       iSet++)
    (iSet->second)->setChangedMember(this, (*projectMap)[name]);

  g_free(name);

  setChanged();
}

/** Open the selected Component.

    \param widget is the view widget

*/
void
Set::openSelectedComponent(GtkWidget *widget)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkTreeStore *store;
  GtkTreeSelection *selection;
  Component *comp;

  if (widget == memberDocView)
      store = memberDocStore;
  else if (widget == memberCatView)
      store = memberCatStore;
  else if (widget == memberSetView)
      store = memberSetStore;
  else
    {
      g_assert_not_reached();
      return;
    }

  // Get the selected Project Component.
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));

  // There should be a selection.
  if (!gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      g_assert_not_reached();
      return;
    }

  gtk_tree_model_get(GTK_TREE_MODEL(store),
		     &iter,
		     Set::ADDRESS, &comp,
		     -1);

  // Open the Component.
  if (comp != NULL)
    comp->showWindow();
}

/***********************************************************************/
/* CALLBACKS */
/***********************************************************************/
/** Insert selected Document into a Set.
 */
static void setInsertDocument_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->insertComponent(QualResObject::DOCUMENT);
}

/** Insert selected Category into a Set.
 */
static void setInsertCategory_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->insertComponent(QualResObject::CATEGORY);
}

/** Insert selected Set into a Set.
 */
static void setInsertSet_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->insertComponent(QualResObject::SET);
}

/** Remove selected Document into a Set.
 */
static void setRemoveDocument_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->removeComponent(QualResObject::DOCUMENT);
}

/** Remove selected Category into a Set.
 */
static void setRemoveCategory_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->removeComponent(QualResObject::CATEGORY);
}

/** Remove selected Set into a Set.
 */
static void setRemoveSet_cb(GtkWidget* widget, gpointer data)
{
  Set *set = (Set *) data;

  set->removeComponent(QualResObject::SET);
}

/** Callback to respond to multiple clicks on Project Components.
 */
static gboolean
setButtonPress_cb(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
  Set *set = (Set *) data;

  // Double click with button 1.  Open the selected Component.
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    set->openSelectedComponent(widget);

  // Continue with normal handling.
  return(FALSE);
}
